import torch
import torch.nn as nn

from .resnext3d_stage import ResStage
from .resnext3d_stem import R2Plus1DStem, ResNeXt3DStem

model_stems = {
    "r2plus1d_stem": R2Plus1DStem,
    "resnext3d_stem": ResNeXt3DStem,
    # For more types of model stem, add them below
}


class FullyConvolutionalLinear(nn.Module):
    def __init__(self, dim_in, num_classes):
        super(FullyConvolutionalLinear, self).__init__()
        # Perform FC in a fully convolutional manner. The FC layer will be
        # initialized with a different std comparing to convolutional layers.
        self.projection = nn.Linear(dim_in, num_classes, bias=True)

    def forward(self, x):
        x = x.view(x.shape[0], -1)
        x = self.projection(x)
        return x


class ResNeXt3D(torch.nn.Module):
    """
    Implementation of:
        1. Conventional `post-activated 3D ResNe(X)t <https://arxiv.org/
        abs/1812.03982>`_.

        2. `Pre-activated 3D ResNe(X)t <https://arxiv.org/abs/1811.12814>`_.
        The model consists of one stem, a number of stages, and one or multiple
        heads that are attached to different blocks in the stage.
    """

    def __init__(
            self,
            input_planes: int = 3,
            skip_transformation_type: str = 'postactivated_shortcut',
            residual_transformation_type: str = 'basic_transformation',
            num_blocks: list = (2, 2, 2, 2),
            stem_name: str = 'resnext3d_stem',
            stem_planes: int = 64,
            stem_temporal_kernel: int = 3,
            stem_spatial_kernel: int = 7,
            stem_maxpool: bool = False,
            stage_planes: int = 64,
            stage_temporal_kernel_basis: list = ([3], [3], [3], [3]),
            temporal_conv_1x1: list = (False, False, False, False),
            stage_temporal_stride: list = (1, 2, 2, 2),
            stage_spatial_stride: list = (1, 2, 2, 2),
            num_groups: int = 1,
            width_per_group: int = 64,
            zero_init_residual_transform: bool = False,
            in_plane: int = 512,
            num_classes: int = 2
    ):
        """
        Args:
            input_planes (int): the channel dimension of the input.
                Normally 3 is used for rgb input.
            skip_transformation_type (str): the type of skip transformation.
                residual_transformation_type (str):
                the type of residual transformation.
            num_blocks (list): list of the number of blocks in stages.
            stem_name (str): name of model stem.
            stem_planes (int): the output dimension
                of the convolution in the model stem.
            stem_temporal_kernel (int): the temporal kernel
                size of the convolution
                in the model stem.
            stem_spatial_kernel (int): the spatial kernel size
                of the convolution in the model stem.
            stem_maxpool (bool): If true, perform max pooling.
            stage_planes (int): the output channel dimension
                of the 1st residual stage
            stage_temporal_kernel_basis (list): Basis of temporal kernel
                sizes for each of the stage.
            temporal_conv_1x1 (bool): Only useful for BottleneckTransformation.
                In a pathaway, if True, do temporal convolution
                in the first 1x1
                Conv3d. Otherwise, do it in the second 3x3 Conv3d.
            stage_temporal_stride (int): the temporal stride of the residual
                transformation.
            stage_spatial_stride (int): the spatial stride of the the residual
                transformation.
            num_groups (int): number of groups for the convolution.
                num_groups = 1 is for standard ResNet like networks, and
                num_groups > 1 is for ResNeXt like networks.
            width_per_group (int): Number of channels per group in 2nd (group)
                conv in the residual transformation in the first stage
            zero_init_residual_transform (bool): if true, the weight of last
                operation, which could be either BatchNorm3D in post-activated
                transformation or Conv3D in pre-activated transformation,
                in the residual transformation is initialized to zero
            pool_size: for fully convolution layer
            in_plane: for fully convolution layer
            num_classes: number of classes
        """

        super().__init__()

        num_stages = len(num_blocks)
        out_planes = [stage_planes * 2 ** i for i in range(num_stages)]
        in_planes = [stem_planes] + out_planes[:-1]
        inner_planes = [
            num_groups * width_per_group * 2 ** i for i in range(num_stages)
        ]

        self.stem = model_stems[stem_name](
            stem_temporal_kernel,
            stem_spatial_kernel,
            input_planes,
            stem_planes,
            stem_maxpool,
        )

        stages = []
        for s in range(num_stages):
            stage = ResStage(
                s + 1,
                # stem is viewed as stage 0, and following stages start from 1
                [in_planes[s]],
                [out_planes[s]],
                [inner_planes[s]],
                [stage_temporal_kernel_basis[s]],
                [temporal_conv_1x1[s]],
                [stage_temporal_stride[s]],
                [stage_spatial_stride[s]],
                [num_blocks[s]],
                [num_groups],
                skip_transformation_type,
                residual_transformation_type,
                disable_pre_activation=(s == 0),
                final_stage=(s == (num_stages - 1)),
            )
            stages.append(stage)

        self.stages = nn.Sequential(*stages)
        self._init_parameter(zero_init_residual_transform)

        self.final_avgpool = nn.AdaptiveAvgPool1d(in_plane)
        self.head_fcl = FullyConvolutionalLinear(
            in_plane, num_classes
        )

    def _init_parameter(self, zero_init_residual_transform):
        for m in self.modules():
            if isinstance(m, nn.Conv3d):
                if (
                        hasattr(m, "final_transform_op")
                        and m.final_transform_op
                        and zero_init_residual_transform
                ):
                    nn.init.constant_(m.weight, 0)
                else:
                    nn.init.kaiming_normal_(
                        m.weight, mode="fan_out", nonlinearity="relu"
                    )
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm3d) and m.affine:
                if (
                        hasattr(m, "final_transform_op")
                        and m.final_transform_op
                        and zero_init_residual_transform
                ):
                    batchnorm_weight = 0.0
                else:
                    batchnorm_weight = 1.0
                nn.init.constant_(m.weight, batchnorm_weight)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, mean=0.0, std=0.01)
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        """
        Args:
            x: Tensor(B, T, W, H, C)
        """
        out = self.stem([x])
        out = self.stages(out)[0]
        out = out.view((out.shape[0], 1, -1))
        out = self.final_avgpool(out)
        out = self.head_fcl(out)

        return out



from .decoder import LinknetDecoder
from ..base import EncoderDecoder
from ..encoders import get_encoder


class Linknet(EncoderDecoder):
    """Linknet_ is a fully convolution neural
    network for fast image semantic segmentation

    Note:
        This implementation by default has 4 skip connections (original - 3).

    Args:
        encoder_name: name of classification model
        (without last dense layers) used as feature
            extractor to build segmentation model.
        encoder_weights: one of ``None`` (random initialization),
         ``imagenet`` (pre-training on ImageNet).
        decoder_use_batchnorm: if ``True``, ``BatchNormalisation``
         layer between ``Conv2D`` and ``Activation`` layers
            is used.
        classes: a number of classes for output
         (output shape - ``(batch, classes, h, w)``).
        activation: activation function used in
        ``.predict(x)`` method for inference.
            One of [``sigmoid``, ``softmax``, callable, None]

    Returns:
        ``torch.nn.Module``: **Linknet**

    .. _Linknet:
        https://arxiv.org/pdf/1707.03718.pdf
    """

    def __init__(
            self,
            encoder_name='resnet34',
            encoder_weights='imagenet',
            decoder_use_batchnorm=True,
            classes=1,
            activation='sigmoid',
    ):
        encoder = get_encoder(
            encoder_name,
            encoder_weights=encoder_weights
        )

        decoder = LinknetDecoder(
            encoder_channels=encoder.out_shapes,
            prefinal_channels=32,
            final_channels=classes,
            use_batchnorm=decoder_use_batchnorm,
        )

        super().__init__(encoder, decoder, activation)

        self.name = 'link-{}'.format(encoder_name)



from sqlalchemy import Table, Column, MetaData, String

meta = MetaData()


def upgrade(migrate_engine):
    conn = migrate_engine.connect()
    trans = conn.begin()

    try:
        meta.bind = conn

        table = Table('project', meta, autoload=True)
        col = Column('sync_folders', String(8000))
        col.create(table)
    except Exception:
        trans.rollback()
        raise
    else:
        trans.commit()


def downgrade(migrate_engine):
    conn = migrate_engine.connect()
    trans = conn.begin()

    try:
        meta.bind = conn

        table = Table('project', meta, autoload=True)
        table.c.continued.drop()
    except Exception:
        trans.rollback()
        raise
    else:
        trans.commit()



import timm
import torch.nn as nn


class Timm(nn.Module):
    def __init__(self, variant, num_classes, pretrained=True, activation=None):
        super().__init__()

        model = timm.create_model(
            variant, pretrained=pretrained,
            num_classes=num_classes)

        self.model = model
        # self.model.fc = nn.Sequential(
        #     LambdaLayer(lambda x: x.unsqueeze_(0)),
        #     nn.AdaptiveAvgPool1d(self.model.fc.in_features),
        #     LambdaLayer(lambda x: x.squeeze_(0).view(x.size(0), -1)),
        #     self.model.fc
        # )

        if callable(activation) or activation is None:
            self.activation = activation
        elif activation == 'softmax':
            self.activation = nn.Softmax(dim=1)
        elif activation == 'sigmoid':
            self.activation = nn.Sigmoid()
        else:
            raise ValueError(
                'Activation should be "sigmoid"/"softmax"/callable/None')

    def forward(self, x):
        res = self.model(x)
        if isinstance(res, tuple):
            res = res[0]
        if self.activation:
            res = self.activation(res)
        return res


__all__ = ['Timm']



import numpy as np
from tqdm import tqdm

import torch
from torch.jit import load
from torch.utils.data import DataLoader, Dataset

from mlcomp.contrib.transform.tta import TtaWrap


def apply_activation(x, activation):
    if not activation:
        return x
    if activation == 'sigmoid':
        return torch.sigmoid(x)
    if activation == 'softmax':
        return torch.softmax(x, 1)
    raise Exception(f'unknown activation = {activation}')


def _infer_batch(model, loader: DataLoader, activation=None):
    for batch in tqdm(loader, total=len(loader)):
        features = batch['features'].cuda()
        logits = model(features)
        p = apply_activation(logits, activation)

        if isinstance(loader.dataset, TtaWrap):
            p = loader.dataset.inverse(p)
        p = p.detach().cpu().numpy()

        yield {'prob': p, 'count': p.shape[0], **batch}


def _infer(model, loader: DataLoader, activation=None):
    pred = []
    for batch in tqdm(loader, total=len(loader)):
        features = batch['features'].cuda()
        logits = model(features)
        p = apply_activation(logits, activation)

        if isinstance(loader.dataset, TtaWrap):
            p = loader.dataset.inverse(p)
        p = p.detach().cpu().numpy()

        pred.append(p)

    pred = np.vstack(pred)
    return pred


def infer(
        x: Dataset,
        file: str,
        batch_size: int = 1,
        batch_mode: bool = False,
        activation=None,
        num_workers: int = 1,
):
    loader = DataLoader(
        x,
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        pin_memory=True
    )
    model = load(file).cuda()
    if batch_mode:
        return _infer_batch(model, loader, activation=activation)

    return _infer(model, loader, activation=activation)


__all__ = ['infer', 'apply_activation']



import numpy as np


def preprocess_input(x, mean=None, std=None, input_space='RGB',
                     input_range=None, **kwargs):
    if input_space == 'BGR':
        x = x[..., ::-1].copy()

    if input_range is not None:
        if x.max() > 1 and input_range[1] == 1:
            x = x / 255.

    if mean is not None:
        mean = np.array(mean)
        x = x - mean

    if std is not None:
        std = np.array(std)
        x = x / std

    return x



from typing import List

from sqlalchemy import func, or_, case

from mlcomp.db.core import PaginatorOptions
from mlcomp.db.enums import TaskStatus, TaskType
from mlcomp.db.models import Project, Dag, Task, ReportTasks, TaskDependence
from mlcomp.db.providers.base import BaseDataProvider
from mlcomp.utils.misc import to_snake, duration_format, now, parse_time


class DagProvider(BaseDataProvider):
    model = Dag

    # noinspection PyMethodMayBeStatic
    def _get_filter(self, query, filter: dict, last_activity):
        if filter.get('project'):
            query = query.filter(Dag.project == filter['project'])
        if filter.get('name'):
            query = query.filter(Dag.name.like(f'%{filter["name"]}%'))
        if filter.get('id'):
            query = query.filter(Dag.id == int(filter['id']))

        if filter.get('id_min'):
            query = query.filter(Dag.id >= filter['id_min'])

        if filter.get('id_max'):
            query = query.filter(Dag.id <= filter['id_max'])

        if filter.get('created_min'):
            created_min = parse_time(filter['created_min'])
            query = query.filter(Dag.created >= created_min)
        if filter.get('created_max'):
            created_max = parse_time(filter['created_max'])
            query = query.filter(Dag.created <= created_max)
        if filter.get('last_activity_min'):
            last_activity_min = parse_time(filter['last_activity_min'])
            query = query.having(last_activity >= last_activity_min)
        if filter.get('last_activity_max'):
            last_activity_max = parse_time(filter['last_activity_max'])
            query = query.having(last_activity <= last_activity_max)
        if filter.get('report'):
            query = query.filter(Dag.report is not None)
        return query

    def get(self, filter: dict, options: PaginatorOptions = None):
        task_status = []
        for e in TaskStatus:
            task_status.append(
                func.sum(
                    case(
                        whens=[(Task.status == e.value, 1)],
                        else_=0
                    ).label(e.name)
                )
            )

        last_activity = func.max(Task.last_activity).label('last_activity')
        funcs = [
            func.count(Task.id).label('task_count'), last_activity,
            func.min(Task.started).label('started'),
            func.max(Task.finished).label('finished')
        ]

        query = self.query(Dag, Project.name, *funcs,
                           *task_status).join(Project)
        query = self._get_filter(query, filter, last_activity)

        status_clauses = []
        for agg, e in zip(task_status, TaskStatus):
            if filter.get('status', {}).get(to_snake(e.name)):
                status_clauses.append(agg > 0)
        if len(status_clauses) > 0:
            query = query.having(or_(*status_clauses))

        query = query.join(Task, isouter=True).group_by(Dag.id, Project.name)
        # Do not include service tasks
        query = query.filter(Task.type < TaskType.Service.value)

        total = query.count()
        paginator = self.paginator(query, options) if options else query
        res = []
        rules = ('-tasks.dag_rel',)
        for dag, \
            project_name, \
            task_count, \
            last_activity, \
            started, \
            finished, \
            *(task_status) in paginator.all():

            items = self.to_dict(dag, rules=rules).items()
            # noinspection PyDictCreation
            r = {
                'task_count': task_count,
                'last_activity': last_activity,
                'started': started,
                'finished': finished,
                **{k: v
                   for k, v in items if k not in ['tasks', 'config']}
            }
            r['project'] = {'name': project_name}

            r['task_statuses'] = [
                {
                    'name': to_snake(e.name),
                    'count': s
                } for e, s in zip(TaskStatus, task_status)
            ]
            r['last_activity'] = self.serializer.serialize_datetime(
                r['last_activity']
            ) if r['last_activity'] else None
            r['started'] = self.serializer.serialize_datetime(r['started']) \
                if r['started'] else None
            r['finished'] = self.serializer.serialize_datetime(
                r['finished']
            ) if r['finished'] else None

            if task_status[TaskStatus.InProgress.value] > 0:
                delta = (now() - started).total_seconds() if started else 0
            elif sum(
                    task_status[TaskStatus.InProgress.value:]
            ) == 0 or not started or not last_activity:
                delta = 0
            else:
                delta = (last_activity - started).total_seconds()

            r['duration'] = duration_format(delta)
            res.append(r)

        if filter.get('report'):
            dag_ids = [r['id'] for r in res]
            tasks_dags = self.query(Task.id, Task.dag). \
                filter(Task.type <= TaskType.Train.value). \
                filter(Task.dag.in_(dag_ids)). \
                all()

            tasks_within_report = self.query(ReportTasks.task). \
                filter(ReportTasks.report == int(filter['report']))

            tasks_within_report = {t[0] for t in tasks_within_report}
            dags_not_full_included = {
                d
                for t, d in tasks_dags if t not in tasks_within_report
            }
            for r in res:
                r['report_full'] = r['id'] not in dags_not_full_included

        projects = self.query(Project.name, Project.id). \
            order_by(Project.id.desc()). \
            limit(20). \
            all()

        projects = [{'name': name, 'id': id} for name, id in projects]
        return {'total': total, 'data': res, 'projects': projects}

    def config(self, id: int):
        return self.by_id(id).config

    def duration(self, t: Task):
        if not t.started:
            return duration_format(0)
        finished = (
            t.finished if t.finished else now()
        )
        delta = (finished - t.started).total_seconds()
        return duration_format(delta)

    def graph(self, id: int):
        tasks = self.query(Task). \
            filter(Task.dag == id). \
            filter(Task.type <= TaskType.Train.value). \
            all()

        task_ids = [t.id for t in tasks]
        dep = self.query(TaskDependence).filter(
            TaskDependence.task_id.in_(task_ids)
        ).all()
        task_by_id = {t.id: t for t in tasks}

        def label(t: Task):
            res = [t.executor]
            if t.status >= TaskStatus.InProgress.value:
                res.append(self.duration(t))
                res.append(
                    f'{t.current_step if t.current_step else ""}/'
                    f'{t.steps if t.steps else ""}'
                )
            return '\n'.join(res)

        nodes = [
            {
                'id': t.id,
                'label': label(t),
                'name': t.name,
                'status': to_snake(TaskStatus(t.status).name)
            } for t in tasks
        ]
        edges = [
            {
                'from': d.depend_id,
                'to': d.task_id,
                'status': to_snake(
                    TaskStatus(task_by_id[d.depend_id].status).name
                )
            } for d in dep
        ]
        return {'nodes': nodes, 'edges': edges}

    def by_project(self, id: int):
        return self.query(Dag).filter(Dag.project == id).all()

    def remove_all(self, ids: List[int]):
        self.query(Dag).filter(Dag.id.in_(ids)).delete(
            synchronize_session=False)
        self.commit()

    def count(self):
        return self.query(Dag).count()


__all__ = ['DagProvider']



from typing import List

import cv2
import torch
import albumentations as A
import numpy as np
from tqdm import tqdm

from mlcomp.contrib.transform.albumentations import ChannelTranspose
from mlcomp.contrib.model import Pretrained
from mlcomp.contrib.model.timm import Timm


def infer(
        files: List[str],
        checkpoint: str,
        model=None,
        class_='Pretrained',
        variant: str = 'resnet34',
        activation: str = 'softmax',
        num_classes: int = 2,
        batch_size: int = 16,
        device: str = 'cuda',
        transforms=None):
    assert activation in ['softmax', 'sigmoid', None]

    if model is None:
        if class_ == 'Pretrained':
            model = Pretrained(variant=variant, num_classes=num_classes)
        elif class_ == 'Timm':
            model = Timm(variant=variant, num_classes=num_classes)
        else:
            raise Exception('unknown model class')
    if transforms is None:
        transforms = A.Compose([
            A.Resize(224, 224),
            A.Normalize(mean=(0.485, 0.456, 0.406),
                        std=(0.229, 0.224, 0.225)
                        ),
            ChannelTranspose()
        ])

    checkpoint = torch.load(checkpoint,
                            map_location=torch.device(device))
    model.load_state_dict(checkpoint['model_state_dict'])
    model = model.to(device)
    model.eval()

    imgs = []
    preds = []

    def predict_batch(imgs):
        if len(imgs) == 0:
            return

        imgs_t = np.array(
            [transforms(image=img)['image'] for img in imgs]
        )
        tensor = torch.from_numpy(imgs_t).to(device)
        pred = model(tensor)
        if activation == 'softmax':
            pred = torch.softmax(pred, dim=1)
        elif activation == 'sigmoid':
            pred = torch.sigmoid(pred)
        pred = pred.detach().cpu().numpy()
        preds.extend(pred.tolist())

        imgs.clear()

    for file in tqdm(files):
        img = cv2.imread(file)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        imgs.append(img)

        if len(imgs) >= batch_size:
            predict_batch(imgs)

    predict_batch(imgs)
    return np.array(preds)


if __name__ == '__main__':
    print(infer([
        '/home/light/projects/deepfake/data/eyes/dfdc_train_part_49/uowiocuqqt.mp4/0.tar/0_0.jpg'
    ], checkpoint='/home/light/mlcomp/models/deepfake/best.pth'))



import datetime

from mlcomp.db.models import Docker
from mlcomp.db.providers.base import BaseDataProvider
from mlcomp.utils.misc import now


class DockerProvider(BaseDataProvider):
    model = Docker

    def get(self, computer: str, name: str):
        return self.query(Docker). \
            filter(Docker.computer == computer). \
            filter(Docker.name == name). \
            one()

    def get_online(self):
        min_activity = now() - datetime.timedelta(seconds=30)
        return self.query(Docker). \
            filter(Docker.last_activity >= min_activity). \
            all()

    def queues_online(self):
        res = []
        for docker in self.get_online():
            res.append((docker.computer,
                        f'{docker.computer}_{docker.name}_supervisor'))
        return res


__all__ = ['DockerProvider']



import torch.nn as nn
import torch.nn.functional as F

from ..base.model import Model


class Conv3x3GNReLU(nn.Module):
    def __init__(self, in_channels, out_channels, upsample=False):
        super().__init__()
        self.upsample = upsample
        self.block = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, (3, 3),
                      stride=1, padding=1, bias=False),
            nn.GroupNorm(32, out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        x = self.block(x)
        if self.upsample:
            x = F.interpolate(x, scale_factor=2, mode='bilinear',
                              align_corners=True)
        return x


class FPNBlock(nn.Module):
    def __init__(self, pyramid_channels, skip_channels):
        super().__init__()
        self.skip_conv = nn.Conv2d(skip_channels, pyramid_channels,
                                   kernel_size=1)

    def forward(self, x):
        x, skip = x

        x = F.interpolate(x, scale_factor=2, mode='nearest')
        skip = self.skip_conv(skip)

        x = x + skip
        return x


class SegmentationBlock(nn.Module):
    def __init__(self, in_channels, out_channels, n_upsamples=0):
        super().__init__()

        blocks = [
            Conv3x3GNReLU(in_channels, out_channels,
                          upsample=bool(n_upsamples))
        ]

        if n_upsamples > 1:
            for _ in range(1, n_upsamples):
                blocks.append(
                    Conv3x3GNReLU(out_channels, out_channels, upsample=True))

        self.block = nn.Sequential(*blocks)

    def forward(self, x):
        return self.block(x)


class FPNDecoder(Model):

    def __init__(
            self,
            encoder_channels,
            pyramid_channels=256,
            segmentation_channels=128,
            final_channels=1,
            dropout=0.2,
    ):
        super().__init__()

        self.conv1 = nn.Conv2d(encoder_channels[0], pyramid_channels,
                               kernel_size=(1, 1))

        self.p4 = FPNBlock(pyramid_channels, encoder_channels[1])
        self.p3 = FPNBlock(pyramid_channels, encoder_channels[2])
        self.p2 = FPNBlock(pyramid_channels, encoder_channels[3])

        self.s5 = SegmentationBlock(pyramid_channels, segmentation_channels,
                                    n_upsamples=3)
        self.s4 = SegmentationBlock(pyramid_channels, segmentation_channels,
                                    n_upsamples=2)
        self.s3 = SegmentationBlock(pyramid_channels, segmentation_channels,
                                    n_upsamples=1)
        self.s2 = SegmentationBlock(pyramid_channels, segmentation_channels,
                                    n_upsamples=0)

        self.dropout = nn.Dropout2d(p=dropout, inplace=True)
        self.final_conv = nn.Conv2d(segmentation_channels, final_channels,
                                    kernel_size=1, padding=0)

        self.initialize()

    def forward(self, x):
        c5, c4, c3, c2, _ = x

        p5 = self.conv1(c5)
        p4 = self.p4([p5, c4])
        p3 = self.p3([p4, c3])
        p2 = self.p2([p3, c2])

        s5 = self.s5(p5)
        s4 = self.s4(p4)
        s3 = self.s3(p3)
        s2 = self.s2(p2)

        x = s5 + s4 + s3 + s2

        x = self.dropout(x)
        x = self.final_conv(x)

        x = F.interpolate(x, scale_factor=4, mode='bilinear',
                          align_corners=True)
        return x



from __future__ import absolute_import, unicode_literals
from celery import Celery
import os
import sys

from mlcomp import REDIS_PASSWORD, REDIS_HOST, REDIS_PORT

sys.path.insert(0, os.path.dirname(__file__))

broker = f'redis://:{REDIS_PASSWORD}@{REDIS_HOST}:{REDIS_PORT}/0'

app = Celery(
    'mlcomp',
    broker=broker,
    backend=broker,
    include=['mlcomp.worker.tasks']
)
__all__ = ['app']



from collections import defaultdict
from typing import List
import os
import json

import albumentations as A

from mlcomp import DATA_FOLDER
from mlcomp.utils.io import yaml_load
from mlcomp.utils.misc import dict_flatten, dict_unflatten


class Config(dict):
    @property
    def data_folder(self):
        return os.path.join(DATA_FOLDER, self['info']['project'])

    @staticmethod
    def from_json(config: str):
        return Config(json.loads(config))

    @staticmethod
    def from_yaml(config: str):
        return Config(yaml_load(config))


def merge_dicts_smart(target: dict, source: dict, sep='/'):
    target_flatten = dict_flatten(target)
    mapping = defaultdict(list)
    hooks = dict()

    for k, v in target_flatten.items():
        parts = k.split(sep)
        for i in range(len(parts) - 1, -1, -1):
            key = sep.join(parts[i:])
            mapping[key].append(k)

            if 0 < i < len(parts) - 1:
                hooks[sep.join(parts[i:-1])] = sep.join(parts[:i + 1])

    for k, v in list(source.items()):
        if isinstance(v, dict):
            source.update(
                {k + sep + kk: vv for kk, vv in dict_flatten(v).items()})

    for k, v in source.items():
        if len(mapping[k]) == 0:
            parts = k.split(sep)
            hook = None
            for i in range(len(parts) - 1, -1, -1):
                h = sep.join(parts[:i])
                if h in hooks:
                    hook = hooks[h] + sep + sep.join(parts[i:])
                    break

            if not hook:
                hook = k

            mapping[k] = [hook]
        assert len(mapping[k]) == 1, f'ambiguous mapping for {k}'
        key = mapping[k][0]
        target_flatten[key] = v

    return dict_unflatten(target_flatten)


def dict_from_list_str(params):
    params = dict(p.split(':') for p in params)
    for k, v in params.items():
        if v.isnumeric():
            if '.' in v:
                params[k] = float(v)
            else:
                params[k] = int(v)
    return params


def parse_albu(configs: List[dict]):
    res = []
    for config in configs:
        assert 'name' in config, f'name is required in {config}'
        config = config.copy()
        name = config.pop('name')
        if name == 'Compose':
            items = config.pop('items')
            aug = A.Compose(parse_albu(items), **config)
        else:
            aug = getattr(A, name)(**config)
        res.append(aug)
    return res


def parse_albu_short(config, always_apply=False):
    if isinstance(config, str):
        if config == 'hflip':
            return A.HorizontalFlip(always_apply=always_apply)
        if config == 'vflip':
            return A.VerticalFlip(always_apply=always_apply)
        if config == 'transpose':
            return A.Transpose(always_apply=always_apply)

        raise Exception(f'Unknwon augmentation {config}')
    assert type(config) == dict
    return parse_albu([config])


__all__ = ['Config', 'merge_dicts_smart', 'parse_albu', 'parse_albu_short',
           'dict_from_list_str']



import os

import migrate.versioning.api as api
from migrate.exceptions import DatabaseAlreadyControlledError

from mlcomp import SA_CONNECTION_STRING


def migrate(connection_string: str = None):
    folder = os.path.dirname(__file__)
    connection_string = connection_string or SA_CONNECTION_STRING
    try:
        api.version_control(url=connection_string, repository=folder)
    except DatabaseAlreadyControlledError:
        pass

    api.upgrade(url=connection_string, repository=folder)


__all__ = ['migrate']



import numpy as np


def mask2rle(img):
    """
    img: numpy array, 1 - mask, 0 - background
    Returns run length as string formatted
    """
    pixels = img.T.flatten()
    pixels = np.concatenate([[0], pixels, [0]])
    runs = np.where(pixels[1:] != pixels[:-1])[0] + 1
    runs[1::2] -= runs[::2]
    return ' '.join(str(x) for x in runs)


def rle2mask(mask_rle, shape):
    """
    mask_rle: run-length as string formatted (start length)
    shape: (width,height) of array to return
    Returns numpy array, 1 - mask, 0 - background

    """
    s = mask_rle.split()
    starts, lengths = [np.asarray(x, dtype=int) for x in
                       (s[0:][::2], s[1:][::2])]
    starts -= 1
    ends = starts + lengths
    img = np.zeros(shape[0] * shape[1], dtype=np.uint8)
    for lo, hi in zip(starts, ends):
        img[lo:hi] = 1
    return img.reshape(shape).T


__all__ = ['mask2rle', 'rle2mask']



import torch.nn as nn

import pretrainedmodels

from mlcomp.contrib.torch.layers import LambdaLayer


class Pretrained(nn.Module):
    def __init__(self, variant, num_classes, pretrained=True, activation=None):
        super().__init__()
        params = {'num_classes': 1000}
        if not pretrained:
            params['pretrained'] = None

        model = pretrainedmodels.__dict__[variant](**params)
        self.model = model
        linear = self.model.last_linear

        if isinstance(linear, nn.Linear):
            self.model.last_linear = nn.Linear(
                model.last_linear.in_features,
                num_classes
            )
            self.model.last_linear.in_channels = linear.in_features
        elif isinstance(linear, nn.Conv2d):
            self.model.last_linear = nn.Conv2d(
                linear.in_channels,
                num_classes,
                kernel_size=linear.kernel_size,
                bias=True
            )
            self.model.last_linear.in_features = linear.in_channels

        self.model.last_linear = nn.Sequential(
            LambdaLayer(lambda x: x.unsqueeze_(0)),
            nn.AdaptiveAvgPool1d(self.model.last_linear.in_channels),
            LambdaLayer(lambda x: x.squeeze_(0).view(x.size(0), -1)),
            self.model.last_linear
        )

        if callable(activation) or activation is None:
            self.activation = activation
        elif activation == 'softmax':
            self.activation = nn.Softmax(dim=1)
        elif activation == 'sigmoid':
            self.activation = nn.Sigmoid()
        else:
            raise ValueError(
                'Activation should be "sigmoid"/"softmax"/callable/None')

    def forward(self, x):
        res = self.model(x)
        if isinstance(res, tuple):
            res = res[0]
        if self.activation:
            res = self.activation(res)
        return res


__all__ = ['Pretrained']



from mlcomp.db.models import Auxiliary
from mlcomp.db.providers.base import BaseDataProvider
from mlcomp.utils.io import yaml_load


class AuxiliaryProvider(BaseDataProvider):
    model = Auxiliary

    def get(self):
        query = self.query(self.model)
        res = dict()
        for r in query.all():
            res[r.name] = yaml_load(r.data)
            res[r.name] = self.serializer(res[r.name])
        return res


__all__ = ['AuxiliaryProvider']



from catalyst.dl import registry
from catalyst.contrib.models.cv.segmentation import (
    Unet, ResnetLinknet, MobileUnet, ResnetUnet, ResnetFPNUnet, ResnetPSPnet,
    FPNUnet, Linknet, PSPnet,
    ResNetLinknet)

from mlcomp.contrib.criterion import RingLoss
from mlcomp.contrib.catalyst.callbacks.inference import InferBestCallback
from mlcomp.contrib.catalyst.optim import OneCycleCosineAnnealLR
from mlcomp.contrib.model.segmentation_model_pytorch import \
            SegmentationModelPytorch
from mlcomp.contrib.model import Pretrained
from mlcomp.contrib.segmentation.deeplabv3.deeplab import DeepLab


def register():
    registry.Criterion(RingLoss)

    registry.Callback(InferBestCallback)

    registry.Scheduler(OneCycleCosineAnnealLR)

    # classification
    registry.Model(Pretrained)

    # segmentation
    registry.Model(Unet)
    registry.Model(ResnetLinknet)
    registry.Model(MobileUnet)
    registry.Model(ResnetUnet)
    registry.Model(ResnetFPNUnet)
    registry.Model(ResnetPSPnet)
    registry.Model(FPNUnet)
    registry.Model(Linknet)
    registry.Model(PSPnet)
    registry.Model(ResNetLinknet)

    registry.Model(SegmentationModelPytorch)
    registry.Model(DeepLab)


__all__ = ['register']



import math
import torch.nn as nn
import torch.utils.model_zoo as model_zoo


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, dilation=1, downsample=None,
                 BatchNorm=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = BatchNorm(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               dilation=dilation, padding=dilation, bias=False)
        self.bn2 = BatchNorm(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = BatchNorm(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride
        self.dilation = dilation

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNet(nn.Module):

    def __init__(self, block, layers, output_stride, BatchNorm,
                 pretrained=True):
        self.inplanes = 64
        super(ResNet, self).__init__()
        blocks = [1, 2, 4]
        if output_stride == 16:
            strides = [1, 2, 2, 1]
            dilations = [1, 1, 1, 2]
        elif output_stride == 8:
            strides = [1, 2, 1, 1]
            dilations = [1, 1, 2, 4]
        else:
            raise NotImplementedError

        # Modules
        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3,
                               bias=False)
        self.bn1 = BatchNorm(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.layer1 = self._make_layer(block, 64, layers[0], stride=strides[0],
                                       dilation=dilations[0],
                                       BatchNorm=BatchNorm)
        self.layer2 = self._make_layer(block, 128, layers[1],
                                       stride=strides[1],
                                       dilation=dilations[1],
                                       BatchNorm=BatchNorm)
        self.layer3 = self._make_layer(block, 256, layers[2],
                                       stride=strides[2],
                                       dilation=dilations[2],
                                       BatchNorm=BatchNorm)
        self.layer4 = self._make_MG_unit(block, 512, blocks=blocks,
                                         stride=strides[3],
                                         dilation=dilations[3],
                                         BatchNorm=BatchNorm)
        self._init_weight()

        if pretrained:
            self._load_pretrained_model()

    def _make_layer(self, block, planes, blocks, stride=1, dilation=1,
                    BatchNorm=None):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                BatchNorm(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(self.inplanes, planes, stride, dilation, downsample,
                  BatchNorm))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, dilation=dilation,
                                BatchNorm=BatchNorm))

        return nn.Sequential(*layers)

    def _make_MG_unit(self, block, planes, blocks, stride=1, dilation=1,
                      BatchNorm=None):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                BatchNorm(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(self.inplanes, planes, stride, dilation=blocks[0] * dilation,
                  downsample=downsample, BatchNorm=BatchNorm))
        self.inplanes = planes * block.expansion
        for i in range(1, len(blocks)):
            layers.append(block(self.inplanes, planes, stride=1,
                                dilation=blocks[i] * dilation,
                                BatchNorm=BatchNorm))

        return nn.Sequential(*layers)

    def forward(self, input):
        x = self.conv1(input)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        low_level_feat = x
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        return x, low_level_feat

    def _init_weight(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _load_pretrained_model(self):
        pretrain_dict = model_zoo.load_url(
            'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth')
        model_dict = {}
        state_dict = self.state_dict()
        for k, v in pretrain_dict.items():
            if k in state_dict:
                model_dict[k] = v
        state_dict.update(model_dict)
        self.load_state_dict(state_dict)


def ResNet101(output_stride, BatchNorm, pretrained=True):
    """Constructs a ResNet-101 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 4, 23, 3], output_stride, BatchNorm,
                   pretrained=pretrained)
    return model


if __name__ == '__main__':
    import torch

    model = ResNet101(BatchNorm=nn.BatchNorm2d, pretrained=True,
                      output_stride=8)
    input = torch.rand(1, 3, 512, 512)
    output, low_level_feat = model(input)
    print(output.size())
    print(low_level_feat.size())



import torch.nn as nn
import torch.nn.functional as F

from catalyst.contrib import registry


@registry.Model
class Net(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(1, 20, 5, 1)
        self.conv2 = nn.Conv2d(20, 50, 5, 1)
        self.fc1 = nn.Linear(4 * 4 * 50, 500)
        self.fc2 = nn.Linear(500, 10)

    def forward(self, x):
        x = F.relu(self.conv1(x))
        x = F.max_pool2d(x, 2, 2)
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, 2, 2)
        x = x.view(-1, 4 * 4 * 50)
        x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)



from sqlalchemy import Table, Column, MetaData, Float

meta = MetaData()


def upgrade(migrate_engine):
    conn = migrate_engine.connect()
    trans = conn.begin()

    try:
        meta.bind = conn

        task = Table('task', meta, autoload=True)
        task_loss = Column('loss', Float)
        task_loss.create(task)
    except Exception:
        trans.rollback()
        raise
    else:
        trans.commit()


def downgrade(migrate_engine):
    conn = migrate_engine.connect()
    trans = conn.begin()

    try:
        meta.bind = conn

        task = Table('task', meta, autoload=True)
        task.c.loss.drop()
    except Exception:
        trans.rollback()
        raise
    else:
        trans.commit()



import sqlalchemy as sa
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship

from mlcomp.db.models.base import Base


class Model(Base):
    __tablename__ = 'model'

    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.String)
    score_local = sa.Column(sa.Float)
    score_public = sa.Column(sa.Float)
    project = sa.Column(sa.Integer, ForeignKey('project.id'))
    dag = sa.Column(sa.Integer, ForeignKey('dag.id'))
    created = sa.Column(sa.DateTime)
    equations = sa.Column(sa.String)
    fold = sa.Column(sa.Integer)

    dag_rel = relationship('Dag', lazy='noload')
    project_rel = relationship('Project', lazy='noload')


__all__ = ['Model']



from os.path import join

import pandas as pd

from mlcomp.utils.config import Config
from mlcomp.contrib.split import stratified_k_fold
from mlcomp.worker.executors import Executor


@Executor.register
class Split(Executor):
    def __init__(
        self,
        variant: str,
        out: str,
        n_splits: int = 5,
        file: str = None,
        label: str = None
    ):
        self.variant = variant
        self.file = file
        self.n_splits = n_splits
        self.out = out
        self.label = label

    def work(self):
        if self.variant == 'frame':
            fold = stratified_k_fold(
                file=self.file, n_splits=self.n_splits, label=self.label
            )
            df = pd.DataFrame({'fold': fold})
            df.to_csv(self.out, index=False)

    @classmethod
    def _from_config(
        cls, executor: dict, config: Config, additional_info: dict
    ):
        file = join(config.data_folder, executor.get('file'))
        return cls(
            variant=executor['variant'],
            out=join(config.data_folder, 'fold.csv'),
            file=file,
            label=executor['label']
        )


__all__ = ['Split']



import pandas as pd
import numpy as np

from torch.utils.data import Dataset


class MnistDataset(Dataset):
    def __init__(
        self,
        file: str,
        fold_csv: str = None,
        fold_number: int = 0,
        is_test: bool = False,
        max_count: int = None,
        transforms=None
    ):
        df = pd.read_csv(file)
        if fold_csv is not None:
            fold = pd.read_csv(fold_csv)
            if is_test:
                df = df[fold['fold'] == fold_number]
            else:
                df = df[fold['fold'] != fold_number]

        if max_count:
            df = df[:max_count]

        if 'label' in df.columns:
            self.y = df['label'].values.astype(np.int)
            df = df.drop(columns='label', axis=0)
        else:
            self.y = None

        self.x = df.values.reshape((-1, 1, 28, 28)).astype(np.float32)
        self.transforms = transforms

    def __len__(self):
        return len(self.x)

    def __getitem__(self, index):
        x = self.x[index]
        if self.transforms:
            x = self.transforms(image=x)['image']
        res = {'features': x}

        if self.y is not None:
            res['targets'] = self.y[index]
        return res



from typing import Tuple

import numpy as np
from torch.nn import CrossEntropyLoss

from torch.nn.functional import cross_entropy
from torch.utils.data import Sampler

from catalyst.core import _State
from catalyst.dl import Callback


class HardNegativeSampler(Sampler, Callback):
    def __init__(
            self,
            data_source,
            name: str,
            count: int,
            batch_size: int = None,
            hard_interval: Tuple[float, float] = (50, 100),
            index_count: int = 1,
            criterion_data: dict = None
    ):
        super().__init__(data_source)
        self.criterion_data = criterion_data
        self.name = name
        self.order = 0
        self.hard_interval = hard_interval
        self.index_count = index_count

        self.sampled = 0
        self.count = count
        self.batch_size = batch_size or count
        self.max_index = len(data_source)
        self.loss = []
        self.indices = []

    def get_loss(self, state: _State, criterion=None, meta: dict = None):
        criterion = criterion or state.criterion
        if isinstance(criterion, dict):
            res = np.zeros(len(state.input['targets']))
            for k, v in criterion.items():
                res += self.get_loss(state, criterion=v,
                                     meta=self.criterion_data[k])
            return res

        output_key = 'logits' if meta is None else meta['output_key']
        input_key = 'targets' if meta is None else meta['input_key']

        if isinstance(criterion, CrossEntropyLoss):
            loss = cross_entropy(state.output[output_key],
                                 state.input[input_key],
                                 reduction='none'
                                 )
            loss = loss.cpu().detach().numpy()
        else:
            loss = []
            for i in range(len(state.input[input_key])):
                v = criterion(
                    state.output[output_key][i:i + 1],
                    state.input[input_key][i:i + 1]
                )
                loss.append(float(v))
            loss = np.array(loss)

        weight = 1 if meta is None else meta['weight']
        return loss * weight

    def __len__(self):
        return self.count

    def random(self, count: int = None):
        count = count or self.batch_size
        replace = count > self.max_index
        indices = []

        for i in range(self.index_count):
            index = np.random.choice(np.arange(0, self.max_index),
                                     replace=replace, size=count)
            indices.append(index)
        return indices

    def sample_batch(self):
        if len(self.loss) > 0:
            percentile_high = np.percentile(self.loss, self.hard_interval[1])
            percentile_low = np.percentile(self.loss, self.hard_interval[0])

            cond = (self.loss >= percentile_low) & (
                    self.loss <= percentile_high)
            hard_samples = np.where(cond)[0]
            indices = [index[hard_samples] for index in self.indices]
        else:
            indices = [[] for _ in range(self.index_count)]

        indices_random = self.random(self.batch_size - len(indices[0]))
        indices_shuffle = np.arange(0, self.batch_size)
        np.random.shuffle(indices_shuffle)

        for i in range(len(indices)):
            indices[i] = np.concatenate([indices[i], indices_random[i]])
            indices[i] = indices[i][indices_shuffle].astype(np.int32)

        if len(indices) == 1:
            return indices[0]
        return zip(indices)

    def __iter__(self):
        while self.sampled < self.count:
            batch = self.sample_batch()
            self.sampled += self.batch_size
            yield from batch

        self.sampled = 0

    def on_batch_end(self, state: _State):
        if state.loader_name != self.name:
            return

        self.loss = self.get_loss(state)
        self.indices = []

        for k in sorted(list(state.input)):
            if 'index_' in k:
                self.indices.append(state.input[k].cpu().detach().numpy())


class HardNegativePairSampler(HardNegativeSampler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, index_count=2, **kwargs)


class HardNegativeTripleSampler(HardNegativeSampler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, index_count=4, **kwargs)


class HardNegativeFourSampler(HardNegativeSampler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, index_count=4, **kwargs)


__all__ = ['HardNegativeSampler', 'HardNegativePairSampler',
           'HardNegativeTripleSampler', 'HardNegativeFourSampler']



import collections
import copy
from datetime import datetime
import re
from typing import List
import os
import pwd
import random

import dateutil
import numpy as np
import signal
import psutil
from subprocess import check_output

first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')


def set_global_seed(seed: int) -> None:
    """
    Sets random seed into PyTorch, TensorFlow, Numpy and Random.

    Args:
        seed: random seed
    """

    try:
        import torch
    except ImportError:
        pass
    else:
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

    random.seed(seed)
    np.random.seed(seed)


def dict_func(objcts: List, func=np.mean):
    if len(objcts) == 0:
        return {}
    first = objcts[0]

    res = dict()
    for k in first:
        k_objcts = [o[k] for o in objcts]
        if isinstance(first[k], dict):
            res[k] = dict_func(k_objcts, func)
        elif isinstance(first[k], list):
            res[k] = [dict_func([o[k][i] for o in objcts], func) for i in
                      range(len(first[k]))]
        else:
            res[k] = func(k_objcts)
    return res


def get_pid(name):
    res = []
    lines = check_output(['ps', '-ef']).decode().split('\n')
    header = lines[0].split()
    for line in lines[1:]:
        if line.strip() == '':
            continue

        parts = line.split()
        parts = parts[:7] + [' '.join(parts[7:])]

        item = {}
        for p, h in zip(parts, header):
            if h == 'CMD:':
                h = 'CMD'

            if h in ['PID', 'PPID']:
                p = int(p)

            item[h] = p
        if 'CMD' in item and name in item['CMD']:
            res.append(item)
    return res


def get_default_network_interface():
    lines = check_output(['route']).decode().split('\n')
    header = lines[1].split()

    for line in lines[2:]:
        parts = line.split()
        item = dict()
        for h, p in zip(header, parts):
            item[h] = p
        if item['Destination'] == 'default' and 'Iface' in item:
            return item['Iface']
    return None


def now():
    return datetime.utcnow()


def merge_dicts(*dicts: dict) -> dict:
    """
    Recursive dict merge.
    Instead of updating only top-level keys,
    ``merge_dicts`` recurses down into dicts nested
    to an arbitrary depth, updating keys.
    Args:
        *dicts: several dictionaries to merge
    Returns:
        dict: deep-merged dictionary
    """
    assert len(dicts) > 1

    dict_ = copy.deepcopy(dicts[0])

    for merge_dict in dicts[1:]:
        merge_dict = merge_dict or {}
        for k, v in merge_dict.items():
            if (
                    k in dict_ and isinstance(dict_[k], dict)
                    and isinstance(merge_dict[k], collections.Mapping)
            ):
                dict_[k] = merge_dicts(dict_[k], merge_dict[k])
            else:
                dict_[k] = merge_dict[k]

    return dict_


def to_snake(name):
    s1 = first_cap_re.sub(r'\1_\2', name)
    return all_cap_re.sub(r'\1_\2', s1).lower()


def log_name(level: int):
    if level == 10:
        return 'DEBUG'
    if level == 20:
        return 'INFO'
    if level == 30:
        return 'WARNING'
    if level == 40:
        return 'ERROR'

    raise Exception('Unknown log level')


def duration_format(delta: float):
    """
    Duration format
    :param delta: seconds
    :return: string representation: 1 days 1 hour 1 min
    """
    if delta < 0:
        delta = f'{int(delta * 1000)} ms'
    elif delta < 60:
        delta = f'{int(delta)} sec'
    elif delta < 3600:
        delta = f'{int(delta / 60)} min {int(delta % 60)} sec'
    elif delta < 3600 * 24:
        hour = int(delta / 3600)
        delta = f'{hour} {"hours" if hour > 1 else "hour"} ' \
                f'{int((delta % 3600) / 60)} min'
    else:
        day = int(delta / (3600 * 24))
        hour = int((delta % (3600 * 24)) / 3600)
        delta = f'{day} {"days" if day > 1 else "day"} {hour}' \
                f' {"hours" if hour > 1 else "hour"} ' \
                f'{int((delta % 3600) / 60)} min'
    return delta


def adapt_db_types(d):
    dic = d.__dict__ if not isinstance(d, dict) else d
    for k in dic:
        if type(dic[k]) in [np.int64]:
            dic[k] = int(dic[k])
        elif type(dic[k]) in [np.float64]:
            dic[k] = float(dic[k])


def dict_flatten(d, parent_key='', sep='/'):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(dict_flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def dict_unflatten(d: dict, sep='/'):
    res = dict()
    for key, value in d.items():
        parts = key.split(sep)
        c = res
        for part in parts[:-1]:
            if part not in c:
                c[part] = dict()
            c = c[part]
        c[parts[-1]] = value

    return res


def memory():
    return map(int, os.popen('free -t -m').readlines()[1].split()[1:4])


def disk(folder: str):
    filesystem, total, used, available, use, mounded_point \
        = os.popen(f'df {folder}').readlines()[1].split()
    total = int(int(total) / 10 ** 6)
    available = int(int(available) / 10 ** 6)
    use = int(use[:-1])
    return total, use, available


def get_username():
    return pwd.getpwuid(os.getuid())[0]


def parse_time(time):
    if not time:
        return None
    if isinstance(time, str):
        return dateutil.parser.parse(time)
    return time


def kill_child_processes(parent_pid, sig=signal.SIGTERM):
    try:
        parent = psutil.Process(parent_pid)
    except psutil.NoSuchProcess:
        return
    children = parent.children(recursive=True)
    for process in children:
        process.send_signal(sig)

    parent.send_signal(sig)


if __name__ == '__main__':
    print(dict_unflatten({'a/b': 10, 'a/e': 20, 'a/c/d': 40}))



import subprocess

from mlcomp.worker.executors import Executor


@Executor.register
class Bash(Executor):
    def __init__(self, command: str, **kwargs):
        super().__init__(**kwargs)

        for k, v in kwargs.items():
            command = command.replace(f'${k}', str(v))

        self.command = command

    def work(self):
        self.info('Opening Process')
        sub_commands = self.command.split('&&')
        for sub in sub_commands:
            self.info('executing '+sub)

            process = subprocess.Popen('exec ' + sub,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE,
                                       shell=True
                                       )
            try:
                self.add_child_process(process.pid)
                self.info('Opening Process. Finished')

                while True:
                    line = process.stdout.readline()
                    if not line:
                        break
                    self.info(line.decode().strip())

                error = []
                while True:
                    line = process.stderr.readline()
                    if not line:
                        break
                    line = line.decode().strip()
                    error.append(line)

                process.communicate()

                if process.returncode != 0:
                    raise Exception('\n'.join(error))
            finally:
                process.kill()


__all__ = ['Bash']



from typing import Tuple

import numpy as np
import cv2


def resize_saving_ratio(img: np.array, size: Tuple[int, int]):
    if not size:
        return img
    if size[0] and img.shape[0] > size[0]:
        k = size[0] / img.shape[0]
        img = cv2.resize(img, (int(k * img.shape[1]), size[0]))
    if size[1] and img.shape[1] > size[1]:
        k = size[1] / img.shape[1]
        img = cv2.resize(img, (size[1], int(k * img.shape[0])))
    return img


__all__ = ['resize_saving_ratio']



from torch.optim.lr_scheduler import CosineAnnealingLR


class OneCycleCosineAnnealLR(CosineAnnealingLR):
    def __init__(self, *args, **kwargs):
        self.start_epoch = None
        self.last_epoch = None
        super().__init__(*args, **kwargs)

    def step(self, epoch=None):
        if self.last_epoch is not None:
            if self.start_epoch is None:
                self.start_epoch = self.last_epoch
                self.last_epoch = 0
                for i in range(len(self.base_lrs)):
                    self.optimizer.param_groups[i]['lr'] = self.base_lrs[0]

            if self.last_epoch >= self.T_max - 1:
                self.start_epoch = self.last_epoch
                self.last_epoch = -1
                for i in range(len(self.base_lrs)):
                    self.optimizer.param_groups[i]['lr'] = self.base_lrs[0]

        super().step(epoch)


__all__ = ['OneCycleCosineAnnealLR']



import sqlalchemy as sa

from mlcomp.db.models.base import Base


class Auxiliary(Base):
    __tablename__ = 'auxiliary'

    name = sa.Column(sa.String, primary_key=True)
    data = sa.Column(sa.String)


__all__ = ['Auxiliary']



import torch
import torch.nn as nn
import torch.nn.functional as F

from ..common.blocks import Conv2dReLU, SCSEModule
from ..base.model import Model


class DecoderBlock(nn.Module):
    def __init__(self, in_channels, out_channels, use_batchnorm=True,
                 attention_type=None):
        super().__init__()
        if attention_type is None:
            self.attention1 = nn.Identity()
            self.attention2 = nn.Identity()
        elif attention_type == 'scse':
            self.attention1 = SCSEModule(in_channels)
            self.attention2 = SCSEModule(out_channels)

        self.block = nn.Sequential(
            Conv2dReLU(in_channels, out_channels, kernel_size=3, padding=1,
                       use_batchnorm=use_batchnorm),
            Conv2dReLU(out_channels, out_channels, kernel_size=3, padding=1,
                       use_batchnorm=use_batchnorm),
        )

    def forward(self, x):
        x, skip = x
        x = F.interpolate(x, scale_factor=2, mode='nearest')
        if skip is not None:
            x = torch.cat([x, skip], dim=1)
            x = self.attention1(x)

        x = self.block(x)
        x = self.attention2(x)
        return x


class CenterBlock(DecoderBlock):

    def forward(self, x):
        return self.block(x)


class UnetDecoder(Model):

    def __init__(
            self,
            encoder_channels,
            decoder_channels=(256, 128, 64, 32, 16),
            final_channels=1,
            use_batchnorm=True,
            center=False,
            attention_type=None
    ):
        super().__init__()

        if center:
            channels = encoder_channels[0]
            self.center = CenterBlock(channels, channels,
                                      use_batchnorm=use_batchnorm)
        else:
            self.center = None

        in_channels = self.compute_channels(encoder_channels, decoder_channels)
        out_channels = decoder_channels

        self.layer1 = DecoderBlock(in_channels[0], out_channels[0],
                                   use_batchnorm=use_batchnorm,
                                   attention_type=attention_type)
        self.layer2 = DecoderBlock(in_channels[1], out_channels[1],
                                   use_batchnorm=use_batchnorm,
                                   attention_type=attention_type)
        self.layer3 = DecoderBlock(in_channels[2], out_channels[2],
                                   use_batchnorm=use_batchnorm,
                                   attention_type=attention_type)
        self.layer4 = DecoderBlock(in_channels[3], out_channels[3],
                                   use_batchnorm=use_batchnorm,
                                   attention_type=attention_type)
        self.layer5 = DecoderBlock(in_channels[4], out_channels[4],
                                   use_batchnorm=use_batchnorm,
                                   attention_type=attention_type)
        self.final_conv = nn.Conv2d(out_channels[4], final_channels,
                                    kernel_size=(1, 1))

        self.initialize()

    def compute_channels(self, encoder_channels, decoder_channels):
        channels = [
            encoder_channels[0] + encoder_channels[1],
            encoder_channels[2] + decoder_channels[0],
            encoder_channels[3] + decoder_channels[1],
            encoder_channels[4] + decoder_channels[2],
            0 + decoder_channels[3],
        ]
        return channels

    def forward(self, x):
        encoder_head = x[0]
        skips = x[1:]

        if self.center:
            encoder_head = self.center(encoder_head)

        x = self.layer1([encoder_head, skips[0]])
        x = self.layer2([x, skips[1]])
        x = self.layer3([x, skips[2]])
        x = self.layer4([x, skips[3]])
        x = self.layer5([x, None])
        x = self.final_conv(x)

        return x



from .f1 import ReportLayoutF1
from .img_classify import ReportLayoutImgClassify
from .info import ReportLayoutInfo
from .metric import ReportLayoutMetric
from .precision_recall import ReportLayoutPrecisionRecall
from .series import ReportLayoutSeries
from .item import ReportLayoutItem

__all__ = [
    'ReportLayoutF1', 'ReportLayoutImgClassify', 'ReportLayoutInfo',
    'ReportLayoutMetric', 'ReportLayoutPrecisionRecall', 'ReportLayoutSeries',
    'ReportLayoutItem'
]



import io
from PIL import Image

import numpy as np

from matplotlib.figure import Figure
import matplotlib.pyplot as plt


def figure_to_binary(figure: Figure, **kwargs):
    buf = io.BytesIO()
    figure.savefig(buf, format='jpg', bbox_inches='tight', **kwargs)
    buf.seek(0)
    content = buf.read()
    buf.close()
    return content


def figure_to_img(figure: Figure, **kwargs):
    buf = io.BytesIO()
    figure.savefig(buf, format='jpg', bbox_inches='tight', **kwargs)
    buf.seek(0)
    im = Image.open(buf)
    im.show()
    buf.close()
    return im


def binary_to_img(b: bytes):
    buf = io.BytesIO()
    buf.write(b)
    buf.seek(0)
    im = Image.open(buf)
    im.show()
    return im


def show_values(pc, fmt='%.2f', **kw):
    """
    Heatmap with text in each cell with matplotlib's pyplot
    Source: https://stackoverflow.com/a/25074150/395857
    By HYRY
    """
    pc.update_scalarmappable()
    ax = pc.axes
    for p, color, value in zip(
            pc.get_paths(), pc.get_facecolors(), pc.get_array()
    ):
        x, y = p.vertices[:-2, :].mean(0)
        if np.all(color[:3] > 0.5):
            color = (0.0, 0.0, 0.0)
        else:
            color = (1.0, 1.0, 1.0)
        ax.text(x, y, fmt % value, ha='center', va='center', color=color, **kw)


def cm2inch(*tupl):
    """
    Specify figure size in centimeter in matplotlib
    Source: https://stackoverflow.com/a/22787457/395857
    By gns-ank
    """
    inch = 2.54
    if type(tupl[0]) == tuple:
        return tuple(i / inch for i in tupl[0])
    else:
        return tuple(i / inch for i in tupl)


def heatmap(
        AUC,
        xlabel,
        ylabel,
        xticklabels,
        yticklabels,
        figure_width=40,
        figure_height=20,
        correct_orientation=False,
        cmap='RdBu'
):
    """
    Inspired by:
    - https://stackoverflow.com/a/16124677/395857
    - https://stackoverflow.com/a/25074150/395857
    """

    # Plot it out
    fig, ax = plt.subplots()
    c = ax.pcolor(
        AUC, edgecolors='k', linestyle='dashed', linewidths=0.2, cmap=cmap
    )

    # put the major ticks at the middle of each cell
    ax.set_yticks(np.arange(AUC.shape[0]) + 0.5, minor=False)
    ax.set_xticks(np.arange(AUC.shape[1]) + 0.5, minor=False)

    # set tick labels
    ax.set_xticklabels(xticklabels, minor=False)
    ax.set_yticklabels(yticklabels, minor=False)

    # set title and x/y labels
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    # Remove last blank column
    plt.xlim((0, AUC.shape[1]))

    # Turn off all the ticks
    ax = plt.gca()
    for t in ax.xaxis.get_major_ticks():
        t.tick1line.set_visible(False)
        t.tick1line.set_visible(False)
    for t in ax.yaxis.get_major_ticks():
        t.tick1line.set_visible(False)
        t.tick1line.set_visible(False)

    # Add color bar
    fig.colorbar(c)

    # Add text in each cell
    show_values(c)

    # Proper orientation (origin at the top left instead of bottom left)
    if correct_orientation:
        ax.invert_yaxis()
        ax.xaxis.tick_top()

    # resize
    fig.set_size_inches(cm2inch(figure_width, figure_height))
    return fig


def plot_classification_report(classification_report, cmap='RdBu'):
    """
    Plot scikit-learn classification report.
    Extension based on https://stackoverflow.com/a/31689645/395857
    """
    lines = classification_report.split('\n')

    classes = []
    plotMat = []
    support = []
    class_names = []
    for line in lines[2:len(lines)]:
        if 'avg' in line or line.strip() == '' or 'accuracy' in line:
            continue
        t = line.strip().split()
        if len(t) < 2:
            continue
        classes.append(t[0])
        v = [float(x) for x in t[1:len(t) - 1]]
        support.append(int(t[-1]))
        class_names.append(t[0])
        plotMat.append(v)

    xlabel = 'Metrics'
    ylabel = 'Classes'
    xticklabels = ['Precision', 'Recall', 'F1-score']
    yticklabels = [
        '{0} ({1})'.format(class_names[idx], sup)
        for idx, sup in enumerate(support)
    ]
    figure_width = 25
    figure_height = len(class_names) + 7
    correct_orientation = False
    return heatmap(
        np.array(plotMat),
        xlabel,
        ylabel,
        xticklabels,
        yticklabels,
        figure_width,
        figure_height,
        correct_orientation,
        cmap=cmap
    )


def show(*imgs, figsize=(20, 8)):
    fig, axes = plt.subplots(1, len(imgs), figsize=figsize)
    if not isinstance(axes, list) and not isinstance(axes, np.ndarray):
        axes = [axes]
    for img, ax in zip(imgs, axes):
        ax.imshow(img)
    plt.show()



import numpy as np
from albumentations import ImageOnlyTransform


class ChannelTranspose(ImageOnlyTransform):
    def get_transform_init_args_names(self):
        return ()

    def get_params_dependent_on_targets(self, params):
        pass

    def __init__(self, axes=(2, 0, 1)):
        super().__init__(always_apply=True)
        self.axes = axes

    def apply(self, img, **params):
        return np.transpose(img, self.axes)


class Ensure4d(ImageOnlyTransform):
    def get_transform_init_args_names(self):
        return ()

    def get_params_dependent_on_targets(self, params):
        pass

    def __init__(self):
        super().__init__(always_apply=True)

    def apply(self, img, **params):
        while len(img.shape) < 4:
            img = img[None]
        return img


__all__ = ['ChannelTranspose', 'Ensure4d']



import torch
import torch.nn.functional as F
from catalyst.contrib.nn.criterion.functional import cosine_distance, \
    batch_all, _EPS


def triplet_loss(
        embeddings: torch.Tensor, labels: torch.Tensor, margin: float = 0.3,
        reduction='mean'
) -> torch.Tensor:
    cosine_dists = cosine_distance(embeddings)
    mask = batch_all(labels)

    anchor_positive_dist = cosine_dists.unsqueeze(2)
    anchor_negative_dist = cosine_dists.unsqueeze(1)
    triplet_loss_value = \
        F.relu(anchor_positive_dist - anchor_negative_dist + margin)
    triplet_loss_value = torch.mul(triplet_loss_value, mask)

    if reduction == 'mean':
        num_positive_triplets = torch.gt(
            triplet_loss_value,
            _EPS).sum().float()
        triplet_loss_value = (
                triplet_loss_value.sum() / (num_positive_triplets + _EPS)
        )
    elif reduction == 'none':
        triplet_loss_value = torch.sum(triplet_loss_value, dim=[1, 2])
    else:
        raise Exception(f'Unknown reduction scheme {reduction}')
    return triplet_loss_value



import os
from os.path import join
import shutil
from pathlib import Path
import sys

import safitty
import torch
from mlcomp.utils.io import yaml_load

from catalyst.dl import Runner
from torch.jit import ScriptModule
import torch.nn as nn

from catalyst.dl.core import Experiment
from catalyst import utils
from catalyst.utils import import_experiment_and_runner

from mlcomp import TASK_FOLDER, MODEL_FOLDER
from mlcomp.db.models import Model
from mlcomp.db.providers import TaskProvider, ModelProvider, \
    ProjectProvider, DagProvider
from mlcomp.utils.misc import now
from mlcomp.utils.config import Config
from mlcomp.worker.executors import Executor


class _ForwardOverrideModel(nn.Module):
    """
    Model that calls specified method instead of forward

    (Workaround, single method tracing is not supported)
    """

    def __init__(self, model, method_name):
        super().__init__()
        self.model = model
        self.method = method_name

    def forward(self, *args, **kwargs):
        args = args[0][self.method]
        if isinstance(args, dict):
            kwargs = args
            args = ()
        return getattr(self.model, self.method)(*args, **kwargs)


class _TracingModelWrapper(nn.Module):
    """
    Wrapper that traces model with batch instead of calling it

    (Workaround, to use native model batch handler)
    """

    def __init__(self, model, method_name):
        super().__init__()
        self.method_name = method_name
        self.model = model
        self.tracing_result: ScriptModule

    def __call__(self, *args, **kwargs):
        method_model = _ForwardOverrideModel(self.model, self.method_name)
        example_inputs = {
            self.method_name: kwargs if len(kwargs) > 0 else args
        }

        # noinspection PyTypeChecker
        self.tracing_result = torch.jit.trace(
            method_model, example_inputs=example_inputs
        )


def trace_model(
    model: Model,
    runner: Runner,
    batch=None,
    method_name: str = "forward",
    mode: str = "eval",
    requires_grad: bool = False,
    opt_level: str = None,
    device: str = "cpu",
    predict_params: dict = None,
) -> ScriptModule:
    """
    Traces model using runner and batch

    Args:
        model: Model to trace
        runner: Model's native runner that was used to train model
        batch: Batch to trace the model
        method_name (str): Model's method name that will be
            used as entrypoint during tracing
        mode (str): Mode for model to trace (``train`` or ``eval``)
        requires_grad (bool): Flag to use grads
        opt_level (str): Apex FP16 init level, optional
        device (str): Torch device
        predict_params (dict): additional parameters for model forward

    Returns:
        (ScriptModule): Traced model
    """
    if batch is None or runner is None:
        raise ValueError("Both batch and runner must be specified.")

    if mode not in ["train", "eval"]:
        raise ValueError(f"Unknown mode '{mode}'. Must be 'eval' or 'train'")

    predict_params = predict_params or {}

    tracer = _TracingModelWrapper(model, method_name)
    if opt_level is not None:
        utils.assert_fp16_available()
        # If traced in AMP we need to initialize the model before calling
        # the jit
        # https://github.com/NVIDIA/apex/issues/303#issuecomment-493142950
        from apex import amp
        model = model.to(device)
        model = amp.initialize(model, optimizers=None, opt_level=opt_level)
        # after fixing this bug https://github.com/pytorch/pytorch/issues/23993
        params = {**predict_params, "check_trace": False}
    else:
        params = predict_params

    getattr(model, mode)()
    utils.set_requires_grad(model, requires_grad=requires_grad)

    _runner_model, _runner_device = runner.model, runner.device

    runner.model, runner.device = tracer, device
    runner.predict_batch(batch, **params)
    result: ScriptModule = tracer.tracing_result

    runner.model, runner.device = _runner_model, _runner_device
    return result


def trace_model_from_checkpoint(logdir, logger, method_name='forward',
                                file='best'):
    config_path = f'{logdir}/configs/_config.json'
    checkpoint_path = f'{logdir}/checkpoints/{file}.pth'
    logger.info('Load config')
    config = safitty.load(config_path)
    if 'distributed_params' in config:
        del config['distributed_params']

    # Get expdir name
    # noinspection SpellCheckingInspection,PyTypeChecker
    # We will use copy of expdir from logs for reproducibility
    expdir_name = config['args']['expdir']
    logger.info(f'expdir_name from args: {expdir_name}')

    sys.path.insert(0, os.path.abspath(join(logdir, '../')))

    expdir_from_logs = os.path.abspath(join(logdir, '../', expdir_name))

    logger.info(f'expdir_from_logs: {expdir_from_logs}')
    logger.info('Import experiment and runner from logdir')

    ExperimentType, RunnerType = \
        import_experiment_and_runner(Path(expdir_from_logs))
    experiment: Experiment = ExperimentType(config)

    logger.info(f'Load model state from checkpoints/{file}.pth')
    model = experiment.get_model(next(iter(experiment.stages)))
    checkpoint = utils.load_checkpoint(checkpoint_path)
    utils.unpack_checkpoint(checkpoint, model=model)

    device = 'cpu'
    stage = list(experiment.stages)[0]
    loader = 0
    mode = 'eval'
    requires_grad = False
    opt_level = None

    runner: RunnerType = RunnerType()
    runner.model, runner.device = model, device

    batch = experiment.get_native_batch(stage, loader)

    logger.info('Tracing')
    traced = trace_model(
        model,
        runner,
        batch,
        method_name=method_name,
        mode=mode,
        requires_grad=requires_grad,
        opt_level=opt_level,
        device=device,
    )

    logger.info('Done')
    return traced


@Executor.register
class ModelAdd(Executor):
    def __init__(
            self,
            name: str,
            project: int,
            fold: int,
            train_task: int = None,
            child_task: int = None,
            file: str = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        self.train_task = train_task
        self.name = name
        self.child_task = child_task
        self.project = project
        self.file = file
        self.fold = fold

    def work(self):
        project = ProjectProvider(self.session).by_id(self.project)

        self.info(f'Task = {self.train_task} child_task: {self.child_task}')

        model = Model(
            created=now(),
            name=self.name,
            project=self.project,
            equations='',
            fold=self.fold
        )

        provider = ModelProvider(self.session)
        if self.train_task:
            task_provider = TaskProvider(self.session)
            dag_provider = DagProvider(self.session)
            task = task_provider.by_id(self.train_task)
            dag = dag_provider.by_id(task.dag)

            task_dir = join(TASK_FOLDER, str(self.child_task or task.id))

            # get log directory
            config = yaml_load(dag.config)
            executor_config = config['executors'][task.executor]
            catalyst_config_file = executor_config['args']['config']
            catalyst_config_file = join(task_dir, catalyst_config_file)
            catalyst_config = yaml_load(file=catalyst_config_file)
            catalyst_logdir = catalyst_config['args']['logdir']

            model.score_local = task.score

            src_log = f'{task_dir}/{catalyst_logdir}'
            models_dir = join(MODEL_FOLDER, project.name)
            os.makedirs(models_dir, exist_ok=True)

            model_path_tmp = f'{src_log}/traced.pth'
            traced = trace_model_from_checkpoint(src_log, self, file=self.file)

            model_path = f'{models_dir}/{model.name}.pth'
            model_weight_path = f'{models_dir}/{model.name}_weight.pth'
            torch.jit.save(traced, model_path_tmp)
            shutil.copy(model_path_tmp, model_path)
            file = self.file = 'best_full'
            shutil.copy(f'{src_log}/checkpoints/{file}.pth',
                        model_weight_path)

        provider.add(model)

    @classmethod
    def _from_config(
            cls, executor: dict, config: Config, additional_info: dict
    ):
        return ModelAdd(
            name=executor['name'],
            project=executor['project'],
            train_task=executor['task'],
            child_task=executor['child_task'],
            fold=executor['fold'],
            file=executor['file']
        )


__all__ = ['ModelAdd', 'trace_model_from_checkpoint']

if __name__ == '__main__':
    import logging
    trace_model_from_checkpoint(
        '/home/ingenix/mlcomp/tasks/64913/log',
        logging
    )


import sqlalchemy as sa
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship, deferred

from mlcomp.db.enums import TaskStatus
from mlcomp.db.models.base import Base


class Task(Base):
    __tablename__ = 'task'

    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.String)
    started = sa.Column(sa.DateTime)
    finished = sa.Column(sa.DateTime)
    last_activity = sa.Column(sa.DateTime)
    computer = sa.Column(sa.String)
    gpu = sa.Column(sa.Integer, default=0)
    gpu_max = sa.Column(sa.Integer, default=0)
    cpu = sa.Column(sa.Integer, default=1)
    executor = sa.Column(sa.String)
    status = sa.Column(sa.Integer, default=TaskStatus.NotRan.value)
    computer_assigned = sa.Column(sa.String, ForeignKey('computer.name'))
    computer_assigned_rel = relationship('Computer', lazy='noload')

    memory = sa.Column(sa.Float, default=0.1)
    steps = sa.Column(sa.Integer, default=1)
    current_step = sa.Column(sa.String)
    dag = sa.Column(sa.Integer, ForeignKey('dag.id'))
    celery_id = sa.Column(sa.String)
    dag_rel = relationship('Dag', lazy='noload')
    debug = sa.Column(sa.Boolean, default=False)
    pid = sa.Column(sa.Integer)
    worker_index = sa.Column(sa.Integer)
    docker_assigned = sa.Column(sa.String)
    type = sa.Column(sa.Integer)
    score = sa.Column(sa.Float)
    report = sa.Column(sa.Integer, ForeignKey('report.id'))
    report_rel = relationship('Report', lazy='noload')
    gpu_assigned = sa.Column(sa.String)
    parent = sa.Column(sa.Integer, ForeignKey('task.id'))
    parent_rel = relationship('Task', lazy='noload')
    loss = sa.Column(sa.Float)

    continued = sa.Column(sa.Boolean, default=False)

    batch_index = sa.Column(sa.Integer)
    batch_total = sa.Column(sa.Integer)
    loader_name = sa.Column(sa.String)
    epoch_duration = sa.Column(sa.Integer)
    epoch_time_remaining = sa.Column(sa.Integer)

    result = deferred(sa.Column(sa.String))
    additional_info = deferred(sa.Column(sa.String))


class TaskDependence(Base):
    __tablename__ = 'task_dependency'

    task_id = sa.Column(sa.Integer, primary_key=True)
    depend_id = sa.Column(sa.Integer, primary_key=True)


class TaskSynced(Base):
    __tablename__ = 'task_synced'

    computer = sa.Column(sa.String, ForeignKey('computer.name'),
                         primary_key=True)
    task = sa.Column(sa.Integer, ForeignKey('task.id'), primary_key=True)


__all__ = ['Task', 'TaskDependence', 'TaskSynced']



import sqlalchemy as sa
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship

from mlcomp.db.models.base import Base


class Step(Base):
    __tablename__ = 'step'

    id = sa.Column(sa.Integer, primary_key=True)
    level = sa.Column(sa.Integer)
    task = sa.Column(sa.Integer, ForeignKey('task.id'))
    started = sa.Column(sa.DateTime)
    finished = sa.Column(sa.DateTime)
    name = sa.Column(sa.String)
    task_rel = relationship('Task', lazy='noload')
    index = sa.Column(sa.Integer)


__all__ = ['Step']



import datetime

from sqlalchemy import func
from sqlalchemy.orm import joinedload

from mlcomp.db.core import PaginatorOptions
from mlcomp.db.enums import DagType
from mlcomp.db.models import Model, Dag, Project, Task
from mlcomp.db.providers.base import BaseDataProvider
from mlcomp.utils.config import Config
from mlcomp.utils.io import yaml_load
from mlcomp.utils.misc import parse_time


class ModelProvider(BaseDataProvider):
    model = Model

    def get(self, filter, options: PaginatorOptions):
        query = self.query(Model). \
            options(joinedload(Model.dag_rel, innerjoin=False)). \
            options(joinedload(Model.project_rel, innerjoin=True))

        if filter.get('project'):
            query = query.filter(Model.project == filter['project'])
        if filter.get('name'):
            query = query.filter(Model.name.like(f'%{filter["name"]}%'))

        if filter.get('created_min'):
            created_min = parse_time(filter['created_min'])
            query = query.filter(Model.created >= created_min)
        if filter.get('created_max'):
            created_max = parse_time(filter['created_max'])
            query = query.filter(Model.created <= created_max)

        total = query.count()
        paginator = self.paginator(query, options) if options else query
        res = []
        models = paginator.all()
        models_projects = set()
        for model in models:
            if not model.project_rel:
                model.project_rel = self.query(Project).filter(
                    Project.id == model.project).one()

            row = self.to_dict(model, rules=('-project_rel.class_names',))
            res.append(row)
            models_projects.add(model.project)

        last_activity = func.max(Task.last_activity).label('last_activity')
        projects = self.query(Project.name, Project.id,
                              last_activity). \
            join(Dag, Dag.project == Project.id, isouter=True). \
            join(Task, isouter=True). \
            group_by(Project.id). \
            order_by(last_activity.desc()). \
            all()
        projects = [{'name': name, 'id': id} for name, id, _ in projects]
        return {'total': total, 'data': res, 'projects': projects}

    def change_dag(self, project: int, name: str, to: int):
        ids = self.query(Model.id). \
            join(Dag). \
            filter(Model.project == project). \
            filter(Dag.name == name). \
            filter(Dag.type == DagType.Pipe.value). \
            all()

        ids = [id[0] for id in ids]

        self.query(Model). \
            filter(Model.id.in_(ids)). \
            update({'dag': to}, synchronize_session=False)
        self.commit()

    def model_start_begin(self, model_id: int):
        model = self.by_id(model_id)

        models_dags = self.query(Dag). \
            filter(Dag.type == DagType.Pipe.value). \
            filter(Dag.project == model.project). \
            order_by(Dag.id.desc()). \
            all()

        used_dag_names = set()
        versions = yaml_load(model.equations)

        res_dags = []
        res_dag = None

        for dag in models_dags:
            if dag.name in used_dag_names:
                continue
            config = Config.from_yaml(dag.config)
            d = {
                'name': dag.name,
                'id': dag.id,
                'pipes': [{
                    'name': p
                } for p in config['pipes']]
            }
            for pipe in d['pipes']:
                pipe['versions'] = versions.get(pipe['name'], [])
                used = [v.get('used', datetime.datetime.min) for v in
                        pipe['versions']]
                pipe['used'] = datetime.datetime.min if len(
                    used) == 0 else max(used)

            d['pipes'] = sorted(d['pipes'], key=lambda x: x['used'],
                                reverse=True)
            for p in d['pipes']:
                del p['used']
                for v in p['versions']:
                    if 'used' in v:
                        del v['used']

            used_dag_names.add(dag.name)
            res_dags.append(d)

            if d['id'] == model.dag:
                res_dag = d

        return {'dags': res_dags, 'dag': res_dag, 'model_id': model_id}


__all__ = ['ModelProvider']



import sys

import sqlalchemy as sa
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship

from mlcomp.db.models.base import Base
from mlcomp.utils.misc import now


class ReportSeries(Base):
    __tablename__ = 'report_series'

    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.String)
    value = sa.Column(sa.Float)
    epoch = sa.Column(sa.Integer)
    time = sa.Column(sa.DateTime)
    task = sa.Column(sa.Integer, ForeignKey('task.id'))
    part = sa.Column(sa.String)
    stage = sa.Column(sa.String)

    task_rel = relationship('Task', lazy='noload')


class ReportImg(Base):
    __tablename__ = 'report_img'

    id = sa.Column(sa.Integer, primary_key=True)
    group = sa.Column(sa.String)
    epoch = sa.Column(sa.Integer)
    task = sa.Column(sa.Integer, ForeignKey('task.id'))
    img = sa.Column(sa.LargeBinary)
    dag = sa.Column(sa.Integer, ForeignKey('dag.id'))
    part = sa.Column(sa.String)
    project = sa.Column(sa.Integer, ForeignKey('project.id'))
    y_pred = sa.Column(sa.Integer)
    y = sa.Column(sa.Integer)
    score = sa.Column(sa.Float)
    size = sa.Column(sa.BigInteger)
    attr1 = sa.Column(sa.Float)
    attr2 = sa.Column(sa.Float)
    attr3 = sa.Column(sa.Float)
    attr4 = sa.Column(sa.Float)
    attr5 = sa.Column(sa.Float)
    attr6 = sa.Column(sa.Float)
    attr7 = sa.Column(sa.Float)
    attr8 = sa.Column(sa.Float)
    attr9 = sa.Column(sa.Float)
    attr1_str = sa.Column(sa.String)
    attr2_str = sa.Column(sa.String)
    attr3_str = sa.Column(sa.String)
    attr4_str = sa.Column(sa.String)
    attr5_str = sa.Column(sa.String)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.size = sys.getsizeof(self.img)


class Report(Base):
    __tablename__ = 'report'

    id = sa.Column(sa.Integer, primary_key=True)
    config = sa.Column(sa.String)
    time = sa.Column(sa.DateTime, default=now())
    name = sa.Column(sa.String)
    project = sa.Column(sa.Integer, ForeignKey('project.id'))
    layout = sa.Column(sa.String)


class ReportTasks(Base):
    __tablename__ = 'report_task'

    id = sa.Column(sa.Integer, primary_key=True)
    report = sa.Column(sa.Integer, ForeignKey('report.id'))
    task = sa.Column(sa.Integer, ForeignKey('task.id'))


class ReportLayout(Base):
    __tablename__ = 'report_layout'

    name = sa.Column(sa.String, primary_key=True)
    content = sa.Column(sa.String)
    last_modified = sa.Column(sa.TIMESTAMP)


__all__ = [
    'ReportSeries', 'ReportImg', 'ReportTasks', 'Report', 'ReportLayout'
]



from os.path import join
from typing import Tuple

import click
import socket
from multiprocessing import cpu_count

import torch

from mlcomp.utils.io import yaml_load, yaml_dump
from mlcomp.contrib.search.grid import grid_cells
from mlcomp.migration.manage import migrate as _migrate
from mlcomp import ROOT_FOLDER, IP, PORT, \
    WORKER_INDEX, SYNC_WITH_THIS_COMPUTER, CAN_PROCESS_TASKS, CONFIG_FOLDER
from mlcomp.db.core import Session
from mlcomp.db.enums import DagType, ComponentType, TaskStatus
from mlcomp.db.models import Computer
from mlcomp.db.providers import \
    ComputerProvider, \
    TaskProvider, \
    StepProvider, \
    ProjectProvider
from mlcomp.report import create_report, check_statuses
from mlcomp.utils.config import merge_dicts_smart, dict_from_list_str
from mlcomp.utils.logging import create_logger
from mlcomp.worker.sync import sync_directed, correct_folders
from mlcomp.worker.tasks import execute_by_id
from mlcomp.utils.misc import memory, disk, get_username, \
    get_default_network_interface, now
from mlcomp.server.back.create_dags import dag_standard, dag_pipe

_session = Session.create_session(key=__name__)


def _dag(config: str, debug: bool = False, control_reqs=True,
         params: Tuple[str] = ()):
    logger = create_logger(_session, name='_dag')
    logger.info('started', ComponentType.Client)

    config_text = open(config, 'r').read()
    config_parsed = yaml_load(config_text)
    params = dict_from_list_str(params)
    config_parsed = merge_dicts_smart(config_parsed, params)
    config_text = yaml_dump(config_parsed)

    logger.info('config parsed', ComponentType.Client)

    type_name = config_parsed['info'].get('type', 'standard')
    if type_name == DagType.Standard.name.lower():
        cells = grid_cells(
            config_parsed['grid']) if 'grid' in config_parsed else [None]
        dags = []
        for cell in cells:
            dag = dag_standard(
                session=_session,
                config=config_parsed,
                debug=debug,
                config_text=config_text,
                config_path=config,
                control_reqs=control_reqs,
                logger=logger,
                component=ComponentType.Client,
                grid_cell=cell
            )
            dags.append(dag)

        return dags

    return [
        dag_pipe(
            session=_session, config=config_parsed, config_text=config_text
        )
    ]


def _create_computer():
    tot_m, used_m, free_m = memory()
    tot_d, used_d, free_d = disk(ROOT_FOLDER)
    computer = Computer(
        name=socket.gethostname(),
        gpu=torch.cuda.device_count(),
        cpu=cpu_count(),
        memory=tot_m,
        ip=IP,
        port=PORT,
        user=get_username(),
        disk=tot_d,
        root_folder=ROOT_FOLDER,
        sync_with_this_computer=SYNC_WITH_THIS_COMPUTER,
        can_process_tasks=CAN_PROCESS_TASKS
    )
    ComputerProvider(_session).create_or_update(computer, 'name')


@click.group()
def main():
    pass


@main.command()
def migrate():
    _migrate()


@main.command()
@click.argument('config')
@click.option('--control_reqs', type=bool, default=True)
@click.option('--params', multiple=True)
def dag(config: str, control_reqs: bool, params):
    check_statuses()
    _dag(config, control_reqs=control_reqs, params=params)


@main.command()
def report():
    create_report()


@main.command()
@click.argument('config')
@click.option('--debug', type=bool, default=True)
@click.option('--params', multiple=True)
def execute(config: str, debug: bool, params):
    check_statuses()
    _create_computer()

    # Fail all InProgress Tasks
    logger = create_logger(_session, __name__)

    provider = TaskProvider(_session)
    step_provider = StepProvider(_session)

    for t in provider.by_status(
            TaskStatus.InProgress, worker_index=WORKER_INDEX
    ):
        step = step_provider.last_for_task(t.id)
        logger.error(
            f'Task Id = {t.id} was in InProgress state '
            f'when another tasks arrived to the same worker',
            ComponentType.Worker, t.computer_assigned, t.id, step
        )
        provider.change_status(t, TaskStatus.Failed)

    # Create dags
    dags = _dag(config, debug, params=params)
    for dag in dags:
        for ids in dag.values():
            for id in ids:
                task = provider.by_id(id)
                task.gpu_assigned = ','.join(
                    [str(i) for i in range(torch.cuda.device_count())])

                provider.commit()
                execute_by_id(id, exit=False)


@main.command()
@click.argument('project')
@click.option('--computer', help='sync computer with all the others')
@click.option(
    '--only_from',
    is_flag=True,
    help='only copy files from the computer to all the others'
)
@click.option(
    '--only_to',
    is_flag=True,
    help='only copy files from all the others to the computer'
)
@click.option(
    '--online',
    is_flag=True,
    help='sync with only online computers'
)
def sync(project: str, computer: str, only_from: bool, only_to: bool,
         online: bool):
    """
    Syncs specified project on this computer with other computers
    """
    check_statuses()
    _create_computer()

    computer = computer or socket.gethostname()
    provider = ComputerProvider(_session)
    project_provider = ProjectProvider(_session)
    computer = provider.by_name(computer)
    computers = provider.all_with_last_activtiy()
    p = project_provider.by_name(project)
    assert p, f'Project={project} is not found'

    sync_folders = yaml_load(p.sync_folders)
    ignore_folders = yaml_load(p.ignore_folders)

    sync_folders = correct_folders(sync_folders, p.name)
    ignore_folders = correct_folders(ignore_folders, p.name)

    if not isinstance(sync_folders, list):
        sync_folders = []
    if not isinstance(ignore_folders, list):
        ignore_folders = []

    folders = [[s, ignore_folders] for s in sync_folders]

    for c in computers:
        if c.name != computer.name:
            if online and (now() - c.last_activity).total_seconds() > 100:
                continue

            if not only_from:
                sync_directed(_session, computer, c, folders)
            if not only_to:
                sync_directed(_session, c, computer, folders)


@main.command()
def init():
    env_path = join(CONFIG_FOLDER, '.env')
    lines = open(env_path).readlines()
    for i in range(len(lines)):
        if 'NCCL_SOCKET_IFNAME' in lines[i]:
            interface = get_default_network_interface()
            if interface:
                lines[i] = f'NCCL_SOCKET_IFNAME={interface}\n'
    open(env_path, 'w').writelines(lines)


if __name__ == '__main__':
    main()



import os
import socket
import time
import traceback
import subprocess
from os.path import join
from typing import List

from mlcomp import FILE_SYNC_INTERVAL
from mlcomp.db.core import Session
from mlcomp.db.enums import ComponentType
from mlcomp.db.models import Computer, TaskSynced
from mlcomp.db.providers import ComputerProvider, \
    TaskSyncedProvider, DockerProvider, ProjectProvider
from mlcomp.utils.logging import create_logger
from mlcomp.utils.misc import now
from mlcomp.utils.io import yaml_load, yaml_dump


def sync_directed(
        session: Session,
        source: Computer,
        target: Computer,
        folders: List
):
    current_computer = socket.gethostname()
    logger = create_logger(session, __name__)
    for folder, excluded in folders:
        end = ' --perms  --chmod=777 --size-only'
        if len(excluded) > 0:
            parts = []
            folder_excluded = False
            for i in range(len(excluded)):
                if excluded[i] == folder:
                    folder_excluded = True
                    break
                if not excluded[i].startswith(folder):
                    continue

                part = os.path.relpath(excluded[i], folder)
                part = f'--exclude {part}'
                parts.append(part)

            if folder_excluded:
                continue

            if len(parts) > 0:
                end += ' ' + ' '.join(parts)

        source_folder = join(source.root_folder, folder)
        target_folder = join(target.root_folder, folder)

        if current_computer == source.name:
            command = f'rsync -vhru -e ' \
                      f'"ssh -p {target.port} -o StrictHostKeyChecking=no" ' \
                      f'{source_folder}/ ' \
                      f'{target.user}@{target.ip}:{target_folder}/ {end}'
        elif current_computer == target.name:
            command = f'rsync -vhru -e ' \
                      f'"ssh -p {source.port} -o StrictHostKeyChecking=no" ' \
                      f'{source.user}@{source.ip}:{source_folder}/ ' \
                      f'{target_folder}/ {end}'
        else:
            command = f'rsync -vhru -e ' \
                      f'"ssh -p {target.port} -o StrictHostKeyChecking=no" ' \
                      f' {source_folder}/ ' \
                      f'{target.user}@{target.ip}:{target_folder}/ {end}'

            command = f'ssh -p {source.port} ' \
                      f'{source.user}@{source.ip} "{command}"'

        logger.info(command, ComponentType.WorkerSupervisor, current_computer)
        try:
            subprocess.check_output(command, shell=True,
                                    stderr=subprocess.STDOUT,
                                    universal_newlines=True)
        except subprocess.CalledProcessError as exc:
            raise Exception(exc.output)


def copy_remote(
        session: Session, computer_from: str, path_from: str, path_to: str
):
    provider = ComputerProvider(session)
    src = provider.by_name(computer_from)
    host = socket.gethostname()
    if host != computer_from:
        c = f'scp -P {src.port} {src.user}@{src.ip}:{path_from} {path_to}'
    else:
        f'cp {path_from} {path_to}'
    subprocess.check_output(c, shell=True)
    return os.path.exists(path_to)


def correct_folders(sync_folders: List[str], project_name: str):
    for i in range(len(sync_folders)):
        s = sync_folders[i]
        parts = s.split('/')
        if parts[0] in ['data', 'models']:
            if len(parts) == 1 or parts[1] != project_name:
                parts[0] = join(parts[0], project_name)
        sync_folders[i] = '/'.join(parts)
    return sync_folders


class FileSync:
    session = Session.create_session(key='FileSync')
    logger = create_logger(session, 'FileSync')

    def process_error(self, e: Exception):
        if Session.sqlalchemy_error(e):
            Session.cleanup('FileSync')
            self.session = Session.create_session(key='FileSync')
            self.logger = create_logger(self.session, 'FileSync')

        hostname = socket.gethostname()
        self.logger.error(
            traceback.format_exc(), ComponentType.WorkerSupervisor,
            hostname
        )

    def sync_manual(self, computer: Computer, provider: ComputerProvider):
        """
        button sync was clicked manually
        """
        if not computer.meta:
            return

        meta = yaml_load(computer.meta)
        if 'manual_sync' not in meta:
            return

        manual_sync = meta['manual_sync']

        project_provider = ProjectProvider(self.session)
        docker_provider = DockerProvider(self.session)

        dockers = docker_provider.get_online()
        project = project_provider.by_id(manual_sync['project'])
        sync_folders = manual_sync['sync_folders']
        ignore_folders = manual_sync['ignore_folders']

        sync_folders = correct_folders(sync_folders, project.name)
        ignore_folders = correct_folders(ignore_folders, project.name)

        if not isinstance(sync_folders, list):
            sync_folders = []
        if not isinstance(ignore_folders, list):
            ignore_folders = []

        for docker in dockers:
            if docker.computer == computer.name:
                continue

            source = provider.by_name(docker.computer)
            folders = [[s, ignore_folders] for s in sync_folders]

            computer.syncing_computer = source.name
            provider.update()

            try:
                sync_directed(
                    self.session,
                    target=computer,
                    source=source,
                    folders=folders
                )
            except Exception as e:
                self.process_error(e)
        del meta['manual_sync']
        computer.meta = yaml_dump(meta)
        provider.update()

    def sync(self):
        hostname = socket.gethostname()
        try:
            provider = ComputerProvider(self.session)
            task_synced_provider = TaskSyncedProvider(self.session)

            computer = provider.by_name(hostname)
            sync_start = now()

            if FILE_SYNC_INTERVAL == 0:
                time.sleep(1)
            else:
                self.sync_manual(computer, provider)

                computers = provider.all_with_last_activtiy()
                computers = [
                    c for c in computers
                    if (now() - c.last_activity).total_seconds() < 10
                ]
                computers_names = {c.name for c in computers}

                for c, project, tasks in task_synced_provider.for_computer(
                        computer.name):
                    if c.sync_with_this_computer:
                        if c.name not in computers_names:
                            self.logger.info(f'Computer = {c.name} '
                                             f'is offline. Can not sync',
                                             ComponentType.WorkerSupervisor,
                                             hostname)
                            continue

                        if c.syncing_computer:
                            continue

                        sync_folders = yaml_load(project.sync_folders)
                        ignore_folders = yaml_load(project.ignore_folders)

                        sync_folders = correct_folders(sync_folders,
                                                       project.name)
                        ignore_folders = correct_folders(ignore_folders,
                                                         project.name)

                        if not isinstance(sync_folders, list):
                            sync_folders = []
                        if not isinstance(ignore_folders, list):
                            ignore_folders = []

                        folders = [[s, ignore_folders] for s in sync_folders]

                        computer.syncing_computer = c.name
                        provider.update()

                        sync_directed(self.session, c, computer, folders)

                    for t in tasks:
                        task_synced_provider.add(
                            TaskSynced(computer=computer.name, task=t.id)
                        )

                    time.sleep(FILE_SYNC_INTERVAL)

            computer.last_synced = sync_start
            computer.syncing_computer = None
            provider.update()
        except Exception as e:
            self.process_error(e)



from collections import defaultdict

import pandas as pd
import numpy as np
from numpy.random.mtrand import RandomState

from sklearn.model_selection import StratifiedKFold


def stratified_group_k_fold(
        label: str,
        group_column: str,
        df: pd.DataFrame = None,
        file: str = None,
        n_splits=5,
        seed: int = 0
):
    random_state = RandomState(seed)

    if file is not None:
        df = pd.read_csv(file)

    labels = defaultdict(set)
    for g, l in zip(df[group_column], df[label]):
        labels[g].add(l)

    group_labels = dict()
    groups = []
    Y = []
    for k, v in labels.items():
        group_labels[k] = random_state.choice(list(v))
        Y.append(group_labels[k])
        groups.append(k)

    index = np.arange(len(group_labels))
    folds = StratifiedKFold(n_splits=n_splits, shuffle=True,
                            random_state=random_state).split(index, Y)

    group_folds = dict()
    for i, (train, val) in enumerate(folds):
        for j in val:
            group_folds[groups[j]] = i

    res = np.zeros(len(df))
    for i, g in enumerate(df[group_column]):
        res[i] = group_folds[g]

    return res.astype(np.int)


def stratified_k_fold(
        label: str, df: pd.DataFrame = None, file: str = None, n_splits=5,
        seed: int = 0
):
    random_state = RandomState(seed)

    if file is not None:
        df = pd.read_csv(file)

    index = np.arange(df.shape[0])
    res = np.zeros(index.shape)
    folds = StratifiedKFold(n_splits=n_splits,
                            random_state=random_state,
                            shuffle=True).split(index, df[label])

    for i, (train, val) in enumerate(folds):
        res[val] = i
    return res.astype(np.int)


__all__ = ['stratified_group_k_fold', 'stratified_k_fold']



from mlcomp.db.models import File, Dag, Project
from mlcomp.db.providers.base import BaseDataProvider


class FileProvider(BaseDataProvider):
    model = File

    def hashs(self, project: int):
        return {
            obj[0]: obj[1]
            for obj in self.query(File.md5, File.id
                                  ).filter(File.project == project).all()
        }

    def remove(self, filter: dict):
        query = self.query(File)
        if filter.get('dag'):
            query = query.filter(File.dag == filter['dag'])
        if filter.get('project'):
            query = query.filter(File.project == filter['project'])
        query.delete(synchronize_session=False)
        self.session.commit()

        query = self.query(Dag)
        if filter.get('dag'):
            query.filter(Dag.id == filter['dag']).update({'file_size': 0})

        if filter.get('project'):
            query.filter(Dag.project == filter['project']). \
                update({'file_size': 0})

        self.session.commit()


__all__ = ['FileProvider']



import ast
import operator
import pickle
from copy import deepcopy
from typing import List

import cv2
import numpy as np
import albumentations as A

from torch.utils.data import Dataset

from mlcomp.db.providers import ModelProvider
from mlcomp.utils.config import parse_albu_short, Config
from mlcomp.utils.torch import infer
from mlcomp.worker.executors import Executor
from mlcomp.contrib.transform.tta import TtaWrap

_OP_MAP = {
    ast.Add: operator.add,
    ast.Sub: operator.sub,
    ast.Mult: operator.mul,
    ast.Invert: operator.neg,
    ast.Div: operator.truediv,
    ast.Pow: operator.pow
}


@Executor.register
class Equation(Executor, ast.NodeVisitor):
    # noinspection PyTypeChecker
    def __init__(
            self,
            model_id: int = None,
            suffix: str = '',
            max_count=None,
            part_size: int = None,
            cache_names: List[str] = (),
            **kwargs
    ):
        super().__init__(**kwargs)

        self.__dict__.update(kwargs)
        self.model_id = model_id
        self.suffix = suffix
        self.max_count = max_count
        self.part_size = part_size
        self.part = None
        self.cache = dict()
        self.cache_names = cache_names
        self.model_name = kwargs.get('model_name')
        self.name = kwargs.get('name')
        if not self.model_name and self.model_id:
            self.model_name = ModelProvider(self.session).by_id(
                self.model_id).name
        if not self.name:
            self.name = self.model_name

        self.suffix = self._solve(suffix)

    def tta(self, x: Dataset, tfms=()):
        x = deepcopy(x)
        transforms = getattr(x, 'transforms')
        if not transforms:
            return x
        assert isinstance(transforms, A.Compose), \
            'only Albumentations transforms are supported'
        index = len(transforms.transforms)
        for i, t in enumerate(transforms.transforms):
            if isinstance(t, A.Normalize):
                index = i
                break
        tfms_albu = []
        for i, t in enumerate(tfms):
            t = parse_albu_short(t, always_apply=True)
            tfms_albu.append(t)
            transforms.transforms.insert(index + i, t)
        return TtaWrap(x, tfms_albu)

    def adjust_part(self, part):
        pass

    def generate_parts(self, count):
        part_size = self.part_size or count
        res = []
        for i in range(0, count, part_size):
            res.append((i, min(count, i + part_size)))
        return res

    def load(self, file: str = None):
        file = file or self.name + f'_{self.suffix}'
        file = f'data/pred/{file}'

        data = pickle.load(open(file, 'rb'))
        data = data[self.part[0]: self.part[1]]

        if isinstance(data, list):
            for row in data:
                if type(row).__module__ == np.__name__:
                    continue
                if isinstance(row, list):
                    for i, c in enumerate(row):
                        row[i] = cv2.imdecode(c, cv2.IMREAD_GRAYSCALE)
            data = np.array(data)

        return data

    def torch(
            self,
            x: Dataset,
            file: str = None,
            batch_size: int = 1,
            activation=None,
            num_workers: int = 1
    ):
        file = (file or self.name) + '.pth'
        file = f'models/{file}'
        return infer(
            x=x,
            file=file,
            batch_size=batch_size,
            activation=activation,
            num_workers=num_workers
        )

    def visit_BinOp(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return _OP_MAP[type(node.op)](left, right)

    def visit_Name(self, node):
        name = node.id
        attr = getattr(self, name, None)
        if attr:
            if isinstance(attr, str):
                res = self._solve(attr)
                if attr in self.cache_names:
                    self.cache[attr] = res
                return res
            return attr
        return str(name)

    def visit_List(self, node):
        return self.get_value(node)

    def visit_Tuple(self, node):
        return self.get_value(node)

    def visit_Num(self, node):
        return node.n

    def visit_Str(self, node):
        return node.s

    def visit_Expr(self, node):
        return self.visit(node.value)

    def visit_pow(self, node):
        return node

    def visit_NameConstant(self, node):
        return node.value

    def get_value(self, node):
        t = type(node)
        if t == ast.NameConstant:
            return node.value
        if t == ast.Name:
            return self.visit_Name(node)
        if t == ast.Str:
            return node.s
        if t == ast.Name:
            return node.id
        if t == ast.Num:
            return node.n
        if t == ast.List:
            res = []
            for e in node.elts:
                res.append(self.get_value(e))
            return res
        if t == ast.Tuple:
            res = []
            for e in node.elts:
                res.append(self.get_value(e))
            return res
        raise Exception(f'Unknown type {t}')

    def visit_Call(self, node):
        name = node.func.id
        f = getattr(self, name)
        if not f:
            raise Exception(f'Equation class does not contain method = {name}')

        args = [self.get_value(a) for a in node.args]
        kwargs = {k.arg: self.get_value(k.value) for k in node.keywords}
        return f(*args, **kwargs)

    def _solve(self, equation):
        if equation is None:
            return None

        equation = str(equation)
        if equation in self.cache:
            return self.cache[equation]

        tree = ast.parse(equation)
        if len(tree.body) == 0:
            return None
        calc = self
        res = calc.visit(tree.body[0])

        return res

    def solve(self, name, parts):
        equation = getattr(self, name)
        for part in parts:
            self.cache = {}
            self.part = part
            self.adjust_part(part)
            res = self._solve(equation)
            if name in self.cache_names:
                self.cache[name] = res
            yield res

    @classmethod
    def _from_config(
            cls, executor: dict, config: Config, additional_info: dict
    ):
        return cls(**executor)


__all__ = ['Equation']



import os
import socket
from collections import OrderedDict
from pathlib import Path
from os.path import join

import torch

from catalyst.utils import set_global_seed, load_checkpoint, \
    import_experiment_and_runner
from catalyst.dl import State, Callback, Runner, CheckpointCallback
from catalyst.dl.callbacks import VerboseLogger
from catalyst.utils.config import parse_args_uargs, dump_environment

from mlcomp import TASK_FOLDER
from mlcomp.db.providers import ReportSeriesProvider, ComputerProvider
from mlcomp.db.report_info import ReportLayoutInfo
from mlcomp.utils.io import yaml_load, yaml_dump
from mlcomp.utils.misc import now
from mlcomp.db.models import ReportSeries
from mlcomp.utils.config import Config, merge_dicts_smart
from mlcomp.worker.executors.base import Executor
from mlcomp.worker.executors.model import trace_model_from_checkpoint
from mlcomp.worker.sync import copy_remote


class Args:
    baselogdir = None
    batch_size = None
    check = False
    config = None
    configs = []
    expdir = None
    logdir = None
    num_epochs = None
    num_workers = None
    resume = None
    seed = 42
    verbose = True

    def _get_kwargs(self):
        return [
            (k, v) for k, v in self.__dict__.items() if not k.startswith('_')
        ]


# noinspection PyTypeChecker
@Executor.register
class Catalyst(Executor, Callback):
    def __init__(
            self,
            args: Args,
            report: ReportLayoutInfo,
            distr_info: dict,
            resume: dict,
            grid_config: dict,
            trace: str,
            params: dict,
            **kwargs
    ):
        super().__init__(**kwargs)

        self.order = 0
        self.resume = resume
        self.distr_info = distr_info
        self.args = args
        self.report = report
        self.experiment = None
        self.runner = None
        self.series_provider = ReportSeriesProvider(self.session)
        self.computer_provider = ComputerProvider(self.session)
        self.grid_config = grid_config
        self.master = True
        self.checkpoint_resume = False
        self.checkpoint_stage_epoch = 0
        self.trace = trace
        self.params = params
        self.last_batch_logged = None
        self.loader_started_time = None
        self.parent = None
        self.loader_step_start = 0

    def get_parent_task(self):
        if self.parent:
            return self.parent
        return self.task

    def callbacks(self):
        result = OrderedDict()
        if self.master:
            result['catalyst'] = self

        return result

    def on_loader_start(self, state: State):
        self.loader_started_time = now()
        self.loader_step_start = state.step

    def on_epoch_start(self, state: State):
        if self.checkpoint_resume and state.stage_epoch == 0:
            state.epoch += 1

        state.stage_epoch = state.stage_epoch + self.checkpoint_stage_epoch
        state.checkpoint_data = {'stage_epoch': state.stage_epoch}
        if self.master:
            self.step.start(1, name=state.stage)

            self.step.start(
                2, name=f'epoch {state.stage_epoch}', index=state.stage_epoch
            )

    def on_batch_start(self, state: State):
        step = state.step - self.loader_step_start
        if self.last_batch_logged and \
                step != state.batch_size * state.loader_len:
            if (now() - self.last_batch_logged).total_seconds() < 10:
                return

        task = self.get_parent_task()
        task.batch_index = int(step / state.batch_size)
        task.batch_total = state.loader_len
        if task.batch_index > task.batch_total:
            state.step = state.batch_size
            task.batch_index = 1
        task.loader_name = state.loader_name

        duration = int((now() - self.loader_started_time).total_seconds())
        task.epoch_duration = duration
        task.epoch_time_remaining = int(duration * (
                task.batch_total / task.batch_index)) - task.epoch_duration
        if state.loss is not None:
            # noinspection PyUnresolvedReferences
            task.loss = float(state.loss['loss'])

        self.task_provider.update()
        self.last_batch_logged = now()

    def on_epoch_end(self, state: State):
        self.step.end(2)
        values = state.metric_manager.epoch_values
        for s in values['valid']:
            for k in values:
                value = values[k][s]

                task_id = self.task.parent or self.task.id
                series = ReportSeries(
                    part=k,
                    name=s,
                    epoch=state.epoch,
                    task=task_id,
                    value=value,
                    time=now(),
                    stage=state.stage
                )
                self.series_provider.add(series)

            if s == self.report.metric.name:
                value = values['valid'][s]
                best = False
                task = self.task
                if task.parent:
                    task = self.task_provider.by_id(task.parent)

                if self.report.metric.minimize:
                    if task.score is None or value < task.score:
                        best = True
                else:
                    if task.score is None or value > task.score:
                        best = True
                if best:
                    task.score = value
                    self.task_provider.update()

    def on_stage_start(self, state: State):
        state.loggers = {'console': VerboseLogger()}

    def on_stage_end(self, state: State):
        self.checkpoint_resume = False
        self.checkpoint_stage_epoch = 0
        self.step.end(1)

    @classmethod
    def _from_config(
            cls, executor: dict, config: Config, additional_info: dict
    ):
        args = Args()
        for k, v in executor['args'].items():
            v = str(v)
            if v in ['False', 'True']:
                v = v == 'True'
            elif v.isnumeric():
                v = int(v)

            setattr(args, k, v)

        assert 'report_config' in additional_info, 'layout was not filled'
        report_config = additional_info['report_config']
        report = ReportLayoutInfo(report_config)
        if len(args.configs) == 0:
            args.configs = [args.config]

        distr_info = additional_info.get('distr_info', {})
        resume = additional_info.get('resume')
        params = executor.get('params', {})
        params.update(additional_info.get('params', {}))

        grid_config = executor.copy()
        grid_config.pop('args', '')

        return cls(
            args=args,
            report=report,
            grid_config=grid_config,
            distr_info=distr_info,
            resume=resume,
            trace=executor.get('trace'),
            params=params
        )

    def set_dist_env(self, config):
        info = self.distr_info
        os.environ['MASTER_ADDR'] = info['master_addr']
        os.environ['MASTER_PORT'] = str(info['master_port'])
        os.environ['WORLD_SIZE'] = str(info['world_size'])

        os.environ['RANK'] = str(info['rank'])
        distributed_params = config.get('distributed_params', {})
        distributed_params['rank'] = 0
        config['distributed_params'] = distributed_params

        if info['rank'] > 0:
            self.master = False

    def parse_args_uargs(self):
        args, config = parse_args_uargs(self.args, [])
        config = merge_dicts_smart(config, self.grid_config)
        config = merge_dicts_smart(config, self.params)

        if self.distr_info:
            self.set_dist_env(config)
        return args, config

    def _checkpoint_fix_config(self, experiment):
        resume = self.resume
        if not resume:
            return
        if experiment.logdir is None:
            return

        checkpoint_dir = join(experiment.logdir, 'checkpoints')
        os.makedirs(checkpoint_dir, exist_ok=True)

        file = 'last_full.pth' if resume.get('load_last') else 'best_full.pth'

        path = join(checkpoint_dir, file)
        computer = socket.gethostname()
        if computer != resume['master_computer']:
            master_computer = self.computer_provider.by_name(
                resume['master_computer'])
            path_from = join(
                master_computer.root_folder, str(resume['master_task_id']),
                experiment.logdir,
                'checkpoints', file
            )
            self.info(
                f'copying checkpoint from: computer = '
                f'{resume["master_computer"]} path_from={path_from} '
                f'path_to={path}'
            )

            success = copy_remote(
                session=self.session,
                computer_from=resume['master_computer'],
                path_from=path_from,
                path_to=path
            )

            if not success:
                self.error(
                    f'copying from '
                    f'{resume["master_computer"]}/'
                    f'{path_from} failed'
                )
            else:
                self.info('checkpoint copied successfully')

        elif self.task.id != resume['master_task_id']:
            path = join(
                TASK_FOLDER, str(resume['master_task_id']), experiment.logdir,
                'checkpoints', file
            )
            self.info(
                f'master_task_id!=task.id, using checkpoint'
                f' from task_id = {resume["master_task_id"]}'
            )

        if not os.path.exists(path):
            self.info(f'no checkpoint at {path}')
            return

        ckpt = load_checkpoint(path)
        stages_config = experiment.stages_config
        for k, v in list(stages_config.items()):
            if k == ckpt['stage']:
                stage_epoch = ckpt['checkpoint_data']['stage_epoch'] + 1

                # if it is the last epoch in the stage
                if stage_epoch == v['state_params']['num_epochs'] \
                        or resume.get('load_best'):
                    del stages_config[k]
                    break

                self.checkpoint_stage_epoch = stage_epoch
                v['state_params']['num_epochs'] -= stage_epoch
                break
            del stages_config[k]

        stage = experiment.stages_config[experiment.stages[0]]
        for k, v in stage['callbacks_params'].items():
            if v.get('callback') == 'CheckpointCallback':
                v['resume'] = path

        self.info(f'found checkpoint at {path}')

    def _checkpoint_fix_callback(self, callbacks: dict):
        def mock(state):
            pass

        for k, c in callbacks.items():
            if not isinstance(c, CheckpointCallback):
                continue

            if c.resume:
                self.checkpoint_resume = True

            if not self.master:
                c.on_epoch_end = mock
                c.on_stage_end = mock
                c.on_batch_start = mock

    def work(self):
        args, config = self.parse_args_uargs()
        set_global_seed(args.seed)

        Experiment, R = import_experiment_and_runner(Path(args.expdir))

        runner_params = config.pop('runner_params', {})

        experiment = Experiment(config)
        runner: Runner = R(**runner_params)

        self.experiment = experiment
        self.runner = runner

        stages = experiment.stages[:]

        if self.task.parent:
            self.parent = self.task_provider.by_id(self.task.parent)

        if self.master:
            task = self.get_parent_task()
            task.steps = len(stages)
            self.task_provider.commit()

        self._checkpoint_fix_config(experiment)

        _get_callbacks = experiment.get_callbacks

        def get_callbacks(stage):
            res = self.callbacks()
            for k, v in _get_callbacks(stage).items():
                res[k] = v

            self._checkpoint_fix_callback(res)
            return res

        experiment.get_callbacks = get_callbacks

        if experiment.logdir is not None:
            dump_environment(config, experiment.logdir, args.configs)

        if self.distr_info:
            info = yaml_load(self.task.additional_info)
            info['resume'] = {
                'master_computer': self.distr_info['master_computer'],
                'master_task_id': self.task.id - self.distr_info['rank'],
                'load_best': True
            }
            self.task.additional_info = yaml_dump(info)
            self.task_provider.commit()

            experiment.stages_config = {
                k: v
                for k, v in experiment.stages_config.items()
                if k == experiment.stages[0]
            }

        runner.run_experiment(experiment, check=args.check)
        if runner.state.exception:
            raise runner.state.exception

        if self.master and self.trace:
            traced = trace_model_from_checkpoint(self.experiment.logdir, self)
            torch.jit.save(traced, self.trace)
        return {'stage': experiment.stages[-1], 'stages': stages}


__all__ = ['Catalyst']



from abc import ABC, abstractmethod

from numpy import isnan

from mlcomp.db.providers import ModelProvider
from mlcomp.worker.executors import Executor
from mlcomp.worker.executors.base.equation import Equation


@Executor.register
class Valid(Equation, ABC):
    def __init__(
            self,
            layout=None,
            fold_number: int = 0,
            plot_count: int = 0,
            **kwargs
    ):
        super().__init__(**kwargs)

        self.layout = layout
        self.fold_number = fold_number
        self.plot_count = plot_count

    @abstractmethod
    def count(self):
        pass

    @abstractmethod
    def plot(self, preds, score):
        pass

    @abstractmethod
    def plot_final(self, score):
        pass

    @abstractmethod
    def score(self, preds):
        pass

    @abstractmethod
    def score_final(self):
        pass

    @abstractmethod
    def create_base(self):
        pass

    @abstractmethod
    def adjust_part(self, part):
        pass

    def key(self):
        return 'y'

    def work(self):
        self.create_base()
        parts = self.generate_parts(self.count())

        for preds in self.solve(self.key(), parts):
            score = self.score(preds)
            if self.layout and self.plot_count > 0:
                self.plot(preds, score)

        score = self.score_final()
        if isnan(score):
            score = -1
        else:
            score = float(score)

        if self.layout:
            self.plot_final(score)

        self.task.score = score
        self.task_provider.update()

        if self.model_id:
            provider = ModelProvider(self.session)
            model = provider.by_id(self.model_id)
            model.score_local = score
            provider.commit()



from functools import wraps

from sqlalchemy import event
from sqlalchemy.orm.exc import StaleDataError

from mlcomp.db.providers import TaskProvider, StepProvider, DagProvider
from mlcomp.db.models import Task, Step, Log, ReportImg, File
from mlcomp.utils.misc import now
from mlcomp.db.core import Session

_session = Session.create_session(key=__name__)


def error_handler(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        global _session
        try:
            f(*args, **kwargs)
        except Exception as e:
            if Session.sqlalchemy_error(e):
                Session.cleanup(key=__name__)
                _session = Session.create_session(key=__name__)
            raise e

    return decorated


@event.listens_for(Task, 'before_update')
@error_handler
def task_before_update(mapper, connection, target):
    target.last_activity = now()
    if target.parent:
        provider = TaskProvider(_session)
        parent = provider.by_id(target.parent)
        if parent is None:
            return

        parent.last_activity = target.last_activity

        try:
            provider.commit()
        except StaleDataError:
            pass


@event.listens_for(Step, 'before_insert')
@event.listens_for(Step, 'before_update')
@error_handler
def step_before_insert_update(mapper, connection, target):
    TaskProvider(_session).update_last_activity(target.task)


@event.listens_for(Log, 'before_insert')
@error_handler
def log_before_insert(mapper, connection, target):
    if target.step is None:
        return
    step = StepProvider(_session).by_id(target.step)
    TaskProvider(_session).update_last_activity(step.task)


@event.listens_for(ReportImg, 'before_insert')
def dag_before_create(mapper, connection, target):
    provider = DagProvider(_session)
    dag = provider.by_id(target.dag)
    dag.img_size += target.size
    provider.commit()


__all__ = [
    'task_before_update', 'step_before_insert_update', 'log_before_insert',
    'dag_before_create'
]



import hashlib
import shutil
import traceback
import requests
import os
import simplejson as json
from collections import OrderedDict
from functools import wraps

from flask import Flask, request, Response, send_from_directory
from flask_cors import CORS
from sqlalchemy.orm import joinedload

import mlcomp.worker.tasks as celery_tasks
from mlcomp import TOKEN, WEB_PORT, WEB_HOST, FLASK_ENV, TMP_FOLDER
from mlcomp.db.enums import TaskStatus, ComponentType
from mlcomp.db.core import PaginatorOptions, Session
from mlcomp.db.providers import ComputerProvider, ProjectProvider, \
    ReportLayoutProvider, ReportProvider, ModelProvider, ReportImgProvider, \
    DagProvider, DagStorageProvider, TaskProvider, LogProvider, StepProvider, \
    FileProvider, AuxiliaryProvider
from mlcomp.db.report_info import ReportLayoutInfo
from mlcomp.server.back.supervisor import register_supervisor
from mlcomp.utils.logging import create_logger
from mlcomp.utils.io import from_module_path, zip_folder
from mlcomp.server.back.create_dags import dag_model_add, dag_model_start
from mlcomp.utils.misc import to_snake, now
from mlcomp.db.models import Model, Report, ReportLayout, Task, File
from mlcomp.utils.io import yaml_load, yaml_dump
from mlcomp.worker.storage import Storage

app = Flask(__name__)
CORS(app)

_read_session = Session.create_session(key='server.read')
_write_session = Session.create_session(key='server.write')

logger = create_logger(_write_session, __name__)
supervisor = None


@app.route('/', defaults={'path': ''}, methods=['GET'])
@app.route('/<path:path>', methods=['GET'])
def send_static(path):
    file = 'index.html'
    if '.' in path:
        file = path

    module_path = from_module_path(__file__, f'../front/dist/mlcomp/')
    return send_from_directory(module_path, file)


def request_data():
    return json.loads(request.data.decode('utf-8'))


def parse_int(args: dict, key: str):
    return int(args[key]) if args.get(key) and args[key].isnumeric() else None


def construct_paginator_options(args: dict, default_sort_column: str):
    return PaginatorOptions(
        sort_column=args.get('sort_column') or default_sort_column,
        sort_descending=args.get('sort_descending', 'true') == 'true',
        page_number=parse_int(args, 'page_number'),
        page_size=parse_int(args, 'page_size'),
    )


def check_auth(token):
    return str(token).strip() == TOKEN


def authenticate():
    return Response(
        'Could not verify your access level for that URL.\n'
        'You have to login with proper credentials', 401,
        {'WWW-Authenticate': 'xBasic'}
    )


def requires_auth(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        token = request.headers.get('Authorization')
        if not token or not check_auth(token):
            return authenticate()
        return f(*args, **kwargs)

    return decorated


def error_handler(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        global _read_session, _write_session, logger

        success = True
        status = 200
        error = ''

        try:
            res = f(*args, **kwargs)
        except Exception as e:
            if Session.sqlalchemy_error(e):
                Session.cleanup('server.read')
                Session.cleanup('server.write')

                _read_session = Session.create_session(key='server.read')
                _write_session = Session.create_session(key='server.write')

                logger = create_logger(_write_session, __name__)

            logger.error(
                f'Requested Url: {request.path}\n\n{traceback.format_exc()}',
                ComponentType.API
            )

            error = traceback.format_exc()
            success = False
            status = 500
            res = None

        res = res or {}
        if isinstance(res, Response):
            return res

        res['success'] = success
        res['error'] = error

        return Response(json.dumps(res, ignore_nan=True), status=status)

    return decorated


@app.route('/api/computers', methods=['POST'])
@requires_auth
@error_handler
def computers():
    data = request_data()
    options = PaginatorOptions(**data['paginator'])
    options.sort_column = 'name'

    provider = ComputerProvider(_read_session)
    return provider.get(data, options)


@app.route('/api/computer_sync_start', methods=['POST'])
@requires_auth
@error_handler
def computer_sync_start():
    provider = ComputerProvider(_read_session)
    return provider.sync_start()


@app.route('/api/computer_sync_end', methods=['POST'])
@requires_auth
@error_handler
def computer_sync_end():
    data = request_data()
    provider = ComputerProvider(_write_session)
    for computer in provider.all():
        if data.get('computer') and data['computer'] != computer.name:
            continue
        meta = yaml_load(computer.meta)
        meta['manual_sync'] = {
            'project': data['id'],
            'sync_folders': yaml_load(data['sync_folders']),
            'ignore_folders': yaml_load(data['ignore_folders']),
        }
        computer.meta = yaml_dump(meta)
    provider.update()


@app.route('/api/projects', methods=['POST'])
@requires_auth
@error_handler
def projects():
    data = request_data()
    options = PaginatorOptions(**data['paginator'])

    provider = ProjectProvider(_read_session)
    res = provider.get(data, options)
    return res


@app.route('/api/project/add', methods=['POST'])
@requires_auth
@error_handler
def project_add():
    data = request_data()

    provider = ProjectProvider(_write_session)
    provider.add_project(
        data['name'],
        yaml_load(data['class_names']),
        yaml_load(data['sync_folders']),
        yaml_load(data['ignore_folders']),
    )


@app.route('/api/project/stop_all_dags', methods=['POST'])
@requires_auth
@error_handler
def stop_all_dags():
    data = request_data()
    provider = TaskProvider(_write_session)
    tasks = provider.by_status(TaskStatus.InProgress,
                               TaskStatus.Queued,
                               TaskStatus.NotRan,
                               project=data['project']
                               )

    for t in tasks:
        info = yaml_load(t.additional_info)
        info['stopped'] = True
        t.additional_info = yaml_dump(info)

    provider.update()
    supervisor.stop_tasks(tasks)


@app.route('/api/project/remove_all_dags', methods=['POST'])
@requires_auth
@error_handler
def remove_all_dags():
    data = request_data()
    provider = DagProvider(_write_session)
    dags = provider.by_project(data['project'])
    provider.remove_all([d.id for d in dags])


@app.route('/api/project/edit', methods=['POST'])
@requires_auth
@error_handler
def project_edit():
    data = request_data()

    provider = ProjectProvider(_write_session)
    provider.edit_project(
        data['name'],
        yaml_load(data['class_names']),
        yaml_load(data['sync_folders']),
        yaml_load(data['ignore_folders']),
    )


@app.route('/api/report/add_start', methods=['POST'])
@requires_auth
@error_handler
def report_add_start():
    return {
        'projects': ProjectProvider(_read_session).get()['data'],
        'layouts': ReportLayoutProvider(_read_session).get()['data']
    }


@app.route('/api/report/add_end', methods=['POST'])
@requires_auth
@error_handler
def report_add_end():
    data = request_data()

    provider = ReportProvider(_write_session)
    layouts = ReportLayoutProvider(_write_session).all()
    layout = layouts[data['layout']]
    report = Report(
        name=data['name'], project=data['project'], config=yaml_dump(layout)
    )
    provider.add(report)


@app.route('/api/layouts', methods=['POST'])
@requires_auth
@error_handler
def report_layouts():
    data = request_data()

    provider = ReportLayoutProvider(_read_session)
    options = PaginatorOptions(**data['paginator'])
    res = provider.get(data, options)
    return res


@app.route('/api/layout/add', methods=['POST'])
@requires_auth
@error_handler
def report_layout_add():
    data = request_data()

    provider = ReportLayoutProvider(_write_session)
    layout = ReportLayout(name=data['name'], content='', last_modified=now())
    provider.add(layout)


@app.route('/api/layout/edit', methods=['POST'])
@requires_auth
@error_handler
def report_layout_edit():
    data = request_data()

    provider = ReportLayoutProvider(_write_session)
    layout = provider.by_name(data['name'])
    layout.last_modified = now()
    if 'content' in data and data['content'] is not None:
        data_loaded = yaml_load(data['content'])
        ReportLayoutInfo(data_loaded)
        layout.content = data['content']
    if 'new_name' in data and data['new_name'] is not None:
        layout.name = data['new_name']

    provider.commit()


@app.route('/api/layout/remove', methods=['POST'])
@requires_auth
@error_handler
def report_layout_remove():
    data = request_data()

    provider = ReportLayoutProvider(_write_session)
    provider.remove(data['name'], key_column='name')


@app.route('/api/model/add', methods=['POST'])
@requires_auth
@error_handler
def model_add():
    data = request_data()
    dag_model_add(_write_session, data)


@app.route('/api/model/remove', methods=['POST'])
@requires_auth
@error_handler
def model_remove():
    data = request_data()
    provider = ModelProvider(_write_session)
    model = provider.by_id(data['id'], joined_load=[Model.project_rel])
    celery_tasks.remove_model(
        _write_session, model.project_rel.name, model.name
    )
    provider.remove(model.id)


@app.route('/api/model/start_begin', methods=['POST'])
@requires_auth
@error_handler
def model_start_begin():
    data = request_data()
    return ModelProvider(_read_session).model_start_begin(data['model_id'])


@app.route('/api/model/start_end', methods=['POST'])
@requires_auth
@error_handler
def model_start_end():
    data = request_data()
    dag_model_start(_write_session, data)


@app.route('/api/img_classify', methods=['POST'])
@requires_auth
@error_handler
def img_classify():
    data = request_data()
    options = PaginatorOptions(**data['paginator'])
    res = ReportImgProvider(_read_session).detail_img_classify(data, options)
    return res


@app.route('/api/img_segment', methods=['POST'])
@requires_auth
@error_handler
def img_segment():
    data = request_data()
    options = PaginatorOptions(**data['paginator'])
    res = ReportImgProvider(_read_session).detail_img_segment(data, options)
    return res


@app.route('/api/config', methods=['POST'])
@requires_auth
@error_handler
def config():
    id = request_data()
    res = DagProvider(_read_session).config(id)
    return {'data': res}


@app.route('/api/graph', methods=['POST'])
@requires_auth
@error_handler
def graph():
    id = request_data()
    res = DagProvider(_read_session).graph(id)
    return res


@app.route('/api/dags', methods=['POST'])
@requires_auth
@error_handler
def dags():
    data = request_data()
    options = PaginatorOptions(**data['paginator'])
    provider = DagProvider(_read_session)
    res = provider.get(data, options)
    return res


@app.route('/api/code', methods=['POST'])
@requires_auth
@error_handler
def code():
    id = request_data()
    res = OrderedDict()
    parents = dict()

    for s, f in DagStorageProvider(_read_session).by_dag(id):
        s.path = s.path.strip()
        parent = os.path.dirname(s.path)
        name = os.path.basename(s.path)
        if name == '':
            continue

        if s.is_dir:
            node = {'name': name, 'children': [], 'id': s.id, 'dag': id,
                    'storage': s.id}
            if not parent:
                res[name] = node
                parents[s.path] = node
            else:
                # noinspection PyUnresolvedReferences
                parents[parent]['children'].append(node)
                parents[os.path.join(parent, name)] = node
        else:
            node = {'name': name, 'id': f.id, 'dag': id, 'storage': s.id}

            try:
                node['content'] = f.content.decode('utf-8')
            except UnicodeDecodeError:
                node['content'] = ''

            if not parent:
                res[name] = node
            else:
                # noinspection PyUnresolvedReferences
                parents[parent]['children'].append(node)

    def sort_key(x):
        if 'children' in x and len(x['children']) > 0:
            return '_' * 5 + x['name']
        return x['name']

    def sort(node: dict):
        if 'children' in node and len(node['children']) > 0:
            node['children'] = sorted(node['children'], key=sort_key)

            for c in node['children']:
                sort(c)

    res = sorted(list(res.values()), key=sort_key)
    for r in res:
        sort(r)

    return {'items': res}


@app.route('/api/update_code', methods=['POST'])
@requires_auth
@error_handler
def update_code():
    data = request_data()
    provider = FileProvider(_write_session)
    file = provider.by_id(data['file_id'])
    content = data['content'].encode('utf-8')
    md5 = hashlib.md5(content).hexdigest()

    if md5 == file.md5:
        return

    if file.dag != data['dag']:
        new_file = File(md5=md5, content=content, project=file.project,
                        dag=data['dag'])
        provider.add(new_file)

        storage = DagStorageProvider(_write_session).by_id(data['storage'])
        storage.file = new_file.id
        provider.commit()
        return {'file': new_file.id}
    else:
        file.content = content
        file.md5 = md5
        provider.commit()
        return {'file': file.id}


@app.route('/api/code_download', methods=['GET'])
@requires_auth
@error_handler
def code_download():
    id = int(request.args['id'])
    storage = Storage(_read_session)
    dag = DagProvider().by_id(id)
    folder = os.path.join(TMP_FOLDER, f'{dag.id}({dag.name})')

    try:
        storage.download_dag(id, folder)

        file_name = f'{dag.id}({dag.name}).zip'
        dst = os.path.join(TMP_FOLDER, file_name)
        zip_folder(folder, dst)
        res = send_from_directory(TMP_FOLDER, file_name)
        os.remove(dst)
        return res
    finally:
        shutil.rmtree(folder, ignore_errors=True)


@app.route('/api/tasks', methods=['POST'])
@requires_auth
@error_handler
def tasks():
    data = request_data()
    options = PaginatorOptions(**data['paginator'])
    provider = TaskProvider(_read_session)
    res = provider.get(data, options)
    return res


@app.route('/api/task/stop', methods=['POST'])
@requires_auth
@error_handler
def task_stop():
    data = request_data()
    provider = TaskProvider(_write_session)
    task = provider.by_id(data['id'], joinedload(Task.dag_rel, innerjoin=True))

    tasks = [task] + provider.children(task.id)
    supervisor.stop_tasks(tasks)


@app.route('/api/task/info', methods=['POST'])
@requires_auth
@error_handler
def task_info():
    data = request_data()
    task = TaskProvider(_read_session).by_id(
        data['id'], joinedload(Task.dag_rel, innerjoin=True)
    )
    return {
        'pid': task.pid,
        'worker_index': task.worker_index,
        'gpu_assigned': task.gpu_assigned,
        'celery_id': task.celery_id,
        'additional_info': task.additional_info or '',
        'result': task.result or '',
        'id': task.id
    }


@app.route('/api/dag/stop', methods=['POST'])
@requires_auth
@error_handler
def dag_stop():
    data = request_data()
    provider = TaskProvider(_write_session)
    id = int(data['id'])
    tasks = provider.by_dag(id)

    supervisor.stop_tasks(tasks)

    dag_provider = DagProvider(_write_session)
    return {'dag': dag_provider.get({'id': id})['data'][0]}


@app.route('/api/dag/start', methods=['POST'])
@requires_auth
@error_handler
def dag_start():
    data = request_data()
    id = int(data['id'])
    supervisor.start_dag(id)


@app.route('/api/auxiliary', methods=['POST'])
@requires_auth
@error_handler
def auxiliary():
    provider = AuxiliaryProvider(_read_session)
    return provider.get()


@app.route('/api/dag/toogle_report', methods=['POST'])
@requires_auth
@error_handler
def dag_toogle_report():
    data = request_data()
    provider = ReportProvider(_write_session)
    if data.get('remove'):
        provider.remove_dag(int(data['id']), int(data['report']))
    else:
        provider.add_dag(int(data['id']), int(data['report']))
    return {'report_full': not data.get('remove')}


@app.route('/api/task/toogle_report', methods=['POST'])
@requires_auth
@error_handler
def task_toogle_report():
    data = request_data()
    provider = ReportProvider(_write_session)
    if data.get('remove'):
        provider.remove_task(int(data['id']), int(data['report']))
    else:
        provider.add_task(int(data['id']), int(data['report']))
    return {'report_full': not data.get('remove')}


@app.route('/api/logs', methods=['POST'])
@requires_auth
@error_handler
def logs():
    provider = LogProvider(_read_session)
    data = request_data()
    options = PaginatorOptions(**data['paginator'])
    res = provider.get(data, options)
    return res


@app.route('/api/reports', methods=['POST'])
@requires_auth
@error_handler
def reports():
    data = request_data()
    options = PaginatorOptions(**data['paginator'])
    provider = ReportProvider(_read_session)
    res = provider.get(data, options)
    return res


@app.route('/api/report', methods=['POST'])
@requires_auth
@error_handler
def report():
    id = request_data()
    provider = ReportProvider(_read_session)
    res = provider.detail(id)
    return res


@app.route('/api/report/update_layout_start', methods=['POST'])
@requires_auth
@error_handler
def report_update_layout_start():
    id = request_data()['id']
    provider = ReportProvider(_write_session)
    res = provider.update_layout_start(id)
    return res


@app.route('/api/report/update_layout_end', methods=['POST'])
@requires_auth
@error_handler
def report_update_layout_end():
    data = request_data()
    provider = ReportProvider(_write_session)
    layout_provider = ReportLayoutProvider(_write_session)
    provider.update_layout_end(
        data['id'], data['layout'], layout_provider.all()
    )
    return provider.detail(data['id'])


@app.route('/api/task/steps', methods=['POST'])
@requires_auth
@error_handler
def steps():
    id = request_data()
    provider = StepProvider(_read_session)
    res = provider.get(id)
    return res


@app.route('/api/token', methods=['POST'])
def token():
    data = request_data()
    if str(data['token']).strip() != TOKEN:
        return Response(
            json.dumps({
                'success': False,
                'reason': 'invalid token'
            }),
            status=401
        )
    return json.dumps({'success': True})


@app.route('/api/project/remove', methods=['POST'])
@requires_auth
@error_handler
def project_remove():
    id = request_data()['id']
    ProjectProvider(_write_session).remove(id)


@app.route('/api/remove_imgs', methods=['POST'])
@requires_auth
@error_handler
def remove_imgs():
    data = request_data()
    provider = ReportImgProvider(_write_session)
    res = provider.remove(data)
    return res


@app.route('/api/remove_files', methods=['POST'])
@requires_auth
@error_handler
def remove_files():
    data = request_data()
    provider = FileProvider(_write_session)
    res = provider.remove(data)
    return res


@app.route('/api/dag/remove', methods=['POST'])
@requires_auth
@error_handler
def dag_remove():
    id = request_data()['id']
    celery_tasks.remove_dag(_write_session, id)

    dag_provider = DagProvider(_write_session)
    dag_provider.remove(id)


@app.route('/api/models', methods=['POST'])
@requires_auth
@error_handler
def models():
    data = request_data()
    options = PaginatorOptions(**data['paginator'])
    provider = ModelProvider(_read_session)
    res = provider.get(data, options)
    return res


@app.route('/api/stop')
@requires_auth
@error_handler
def stop():
    pass


def shutdown_server():
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()


@app.route('/api/shutdown', methods=['POST'])
@requires_auth
@error_handler
def shutdown():
    shutdown_server()
    return 'Server shutting down...'


def start_server():
    global supervisor
    if os.environ.get('WERKZEUG_RUN_MAIN') != 'true':
        logger.info(f'Server TOKEN = {TOKEN}', ComponentType.API)
        supervisor = register_supervisor()

    app.run(debug=FLASK_ENV == 'development', port=WEB_PORT, host=WEB_HOST)


def stop_server():
    requests.post(
        f'http://localhost:{WEB_PORT}/api/shutdown',
        headers={'Authorization': TOKEN}
    )



import ast
from os.path import join
from numbers import Number
from collections import defaultdict
import os
from typing import Callable, Dict

import tifffile
import numpy as np
import pandas as pd
import cv2

from torch.utils.data import Dataset


def isotropically_resize_image(img, size, resample=cv2.INTER_AREA):
    h, w = img.shape[:2]
    if w > h:
        h = h * size // w
        w = size
    else:
        w = w * size // h
        h = size

    resized = cv2.resize(img, (w, h), interpolation=resample)
    return resized


def make_square_image(img):
    h, w = img.shape[:2]
    size = max(h, w)
    t = 0
    b = size - h
    l = 0
    r = size - w
    return cv2.copyMakeBorder(img, t, b, l, r, cv2.BORDER_CONSTANT, value=0)


class ImageDataset(Dataset):
    def __init__(
            self,
            *,
            img_folder: str,
            fold_csv: str = None,
            fold: int = None,
            is_test: bool = False,
            gray_scale: bool = False,
            num_classes=2,
            max_count=None,
            meta_cols=(),
            transforms=None,
            postprocess_func: Callable[[Dict], Dict] = None,
            include_image_orig=False
    ):
        self.img_folder = img_folder

        if fold_csv:
            df = pd.read_csv(fold_csv)
            if fold is not None:
                if is_test:
                    self.data = df[df['fold'] == fold]
                else:
                    self.data = df[df['fold'] != fold]
            else:
                self.data = df
        else:
            self.data = pd.DataFrame(
                {'image': os.listdir(img_folder)}).sort_values(by='image')

        self.data = self.data.to_dict(orient='row')
        if max_count is not None:
            self.apply_max_count(max_count)

        for row in self.data:
            self.preprocess_row(row)

        self.transforms = transforms
        self.gray_scale = gray_scale
        self.num_classes = num_classes
        self.meta_cols = meta_cols
        self.postprocess_func = postprocess_func
        self.include_image_orig = include_image_orig

    def apply_max_count(self, max_count):
        if isinstance(max_count, Number):
            self.data = self.data[:max_count]
        else:
            data = defaultdict(list)
            for row in self.data:
                data[row['label']].append(row)
            min_index = np.argmin(max_count)
            min_count = len(data[min_index])
            for k, v in data.items():
                count = int(min_count * (max_count[k] / max_count[min_index]))
                data[k] = data[k][:count]

            self.data = [v for i in range(len(data)) for v in data[i]]

    def preprocess_row(self, row: dict):
        row['image'] = join(self.img_folder, row['image'])

    def __len__(self):
        return len(self.data)

    def _get_item_before_transform(self, row: dict, item: dict):
        pass

    def _get_item_after_transform(self, row: dict,
                                  transformed: dict,
                                  res: dict):
        if 'label' in row:
            res['targets'] = ast.literal_eval(str(row['label']))
            if isinstance(res['targets'], list):
                res['targets'] = np.array(res['targets'], dtype=np.float32)

    def __getitem__(self, index):
        row = self.data[index]
        image = self.read_image_file(row['image'], self.gray_scale)
        image = isotropically_resize_image(image, 256)
        image = make_square_image(image)
        item = {'image': image}

        self._get_item_before_transform(row, item)

        if self.transforms:
            item = self.transforms(**item)
        if self.gray_scale:
            item['image'] = np.expand_dims(item['image'], axis=0)
        res = {
            'features': item['image'].astype(np.float32),
            'image_file': row['image']
        }
        if self.include_image_orig:
            res['image'] = image

        for c in self.meta_cols:
            res[c] = row[c]

        self._get_item_after_transform(row, item, res)
        if self.postprocess_func:
            res = self.postprocess_func(res)
        return res

    @staticmethod
    def read_image_file(path: str, gray_scale=False):
        if not os.path.exists(path):
            raise Exception(f'Image at path {path} does not exist')

        if path.endswith('.tiff') and not gray_scale:
            return tifffile.imread(path)
        elif path.endswith('.npy'):
            return np.load(path)
        else:
            if gray_scale:
                img = cv2.imread(path, cv2.IMREAD_GRAYSCALE)
                assert img is not None, \
                    f'Image at path {path} is empty'
                return img.astype(np.uint8)
            else:
                img = cv2.imread(path)
                assert img is not None, \
                    f'Image at path {path} is empty'
                return cv2.cvtColor(img, cv2.COLOR_BGR2RGB)


__all__ = ['ImageDataset']
